"""Static Routes Classes."""

from fmcapi.api_objects.apiclasstemplate import APIClassTemplate
from .devicerecords import DeviceRecords
import logging


class StaticRoutes(APIClassTemplate):
    """The StaticRoutes Object in the FMC."""

    VALID_JSON_DATA = [
        "id",
        "name",
        "type",
        "continentId",
        "continents",
        "countries",
        "continentUUID",
    ]
    VALID_FOR_KWARGS = VALID_JSON_DATA + []
    PREFIX_URL = "/devices/devicerecords"
    URL_SUFFIX = None

    def __init__(self, fmc, **kwargs):
        """
        Initialize StaticRoutes object.

        :param fmc (object): FMC object
        :param **kwargs: Any other values passed during instantiation.
        :return: None
        """
        super().__init__(fmc, **kwargs)
        logging.debug("In __init__() for StaticRoutes class.")
        self.parse_kwargs(**kwargs)

    def device(self, device_name):
        """
        Associate device to this static route.

        :param device_name: (str) Name of device.
        :return: None
        """
        logging.debug("In device() for StaticRoutes class.")
        device1 = DeviceRecords(fmc=self.fmc)
        device1.get(name=device_name)
        if "id" in device1.__dict__:
            self.device_id = device1.id
            self.URL = f"{self.fmc.configuration_url}{self.PREFIX_URL}/{self.device_id}/routing/staticroutes"
            self.device_added_to_url = True
        else:
            logging.warning(
                f'Device "{device_name}" not found.  Cannot set up device for physicalInterface.'
            )

    def post(self):
        """POST method for API for StaticRoutes not supported."""
        logging.info("POST method for API for StaticRoutes not supported.")
        pass

    def put(self):
        """PUT method for API for StaticRoutes not supported."""
        logging.info("PUT method for API for StaticRoutes not supported.")
        pass

    def delete(self):
        """DELETE method for API for StaticRoutes not supported."""
        logging.info("DELETE method for API for StaticRoutes not supported.")
        pass



"""Hit Counts Class."""

from fmcapi.api_objects.apiclasstemplate import APIClassTemplate
from .accesspolicies import AccessPolicies
from .accessrules import AccessRules
from fmcapi.api_objects.device_services.devicerecords import DeviceRecords
from .prefilterpolicies import PreFilterPolicies
import logging


class HitCounts(APIClassTemplate):
    """The HitCounts Object in the FMC."""

    VALID_JSON_DATA = []
    VALID_FOR_KWARGS = VALID_JSON_DATA + [
        "acp_id",
        "acp_name",
        "device_id",
        "device_name",
        "prefilter_id",
        "prefilter_name",
        "fetchZeroHitcount",
        "limit",
        "lastFetchTimeStamp",
        "hitCount",
        "firstHitTimeStamp",
        "lastHitTimeStamp",
        "version",
        "policy",
        "offset",
    ]
    ACP_PREFIX_URL = "/policy/accesspolicies"
    PREFILTER_PREFIX_URL = "/policy/prefilterpolicies"
    REQUIRED_FOR_PUT = ["acp_id", "device_id"]
    REQUIRED_FOR_DELETE = ["acp_id", "device_id"]
    REQUIRED_FOR_GET = ["device_id"]
    FIRST_SUPPORTED_FMC_VERSION = "6.4"

    @property
    def URL_SUFFIX(self):
        """Add the URL suffixes for filter."""
        filter_init = '?filter="'
        filter_string = filter_init

        self.URL = self.URL.split("?")[0]

        if self.device_id:
            filter_string += f"deviceId:{self.device_id};"
        if self.acp_rule_ids:
            filter_string += f"ids:{','.join(self.acp_rule_ids)};"
        if self.fetchZeroHitCount:
            filter_string += f"fetchZeroHitCount:{self._fetchZeroHitCount};"

        if filter_string == filter_init:
            filter_string += '"'
        filter_string = f'{filter_string[:-1]}"&expanded=true'

        if "limit" in self.__dict__:
            filter_string += f"&limit={self.limit}"
        if "offset" in self.__dict__:
            filter_string += f"&offset={self.offset}"
        return filter_string

    @property
    def fetchZeroHitCount(self):
        """Getter for fetchZeroHitCount."""
        return self._fetchZeroHitCount

    @fetchZeroHitCount.setter
    def fetchZeroHitCount(self, value=False):
        """Setter for fetchZeroHitCount."""
        self._fetchZeroHitCount = value
        # Rebuild the URL with possible new information
        self.URL = self.URL.split("?")[0]
        self.URL = f"{self.URL}{self.URL_SUFFIX}"

    def __init__(self, fmc, **kwargs):
        """
        Initialize HitCounts object.

        :param fmc (object): FMC object
        :param **kwargs: Any other values passed during instantiation.
        :return: None
        """
        logging.debug("In __init__() for HitCounts class.")
        self.device_id = None
        self.prefilter_id = None
        self.acp_id = None
        self.acp_rule_ids = []
        self.fetchZeroHitCount = False
        super().__init__(fmc, **kwargs)
        self.parse_kwargs(**kwargs)
        self.type = "HitCount"
        self.URL = f"{self.URL}"

    def parse_kwargs(self, **kwargs):
        """
        Parse the kwargs and set self variables to match.

        :return: None
        """
        super().parse_kwargs(**kwargs)
        logging.debug("In parse_kwargs() for HitCounts class.")
        if "acp_id" in kwargs:
            self.acp(acp_id=kwargs["acp_id"])
        if "acp_name" in kwargs:
            self.acp(name=kwargs["acp_name"])
        if "device_id" in kwargs:
            self.device(device_id=kwargs["device_id"])
        if "device_name" in kwargs:
            self.device(name=kwargs["device_name"])
        if "prefilter_id" in kwargs:
            self.prefilter_policy(prefilter_policy_id=kwargs["prefilter_id"])
        if "prefilter_name" in kwargs:
            self.prefilter_policy(name=kwargs["prefilter_name"])

    def device(self, name="", device_id=""):
        """
        Associate a device with this HitCount.

        Either the 'name' or the 'device_id' is required, not both.
        :param name: (str) Name of device.
        :param device_id: (str) UUID of device.
        :return: None
        """
        logging.debug("In device() for HitCounts class")
        if device_id != "":
            self.device_id = device_id
        elif name != "":
            device1 = DeviceRecords(fmc=self.fmc)
            device1.get(name=name)
            if "id" in device1.__dict__:
                self.device_id = device1.id
            else:
                logging.warning(
                    f'Device "{name}" not found.  Cannot configure device for HitCounts.'
                )
        else:
            logging.error("No device name or id was provided.")

        # Rebuild the URL with possible new information
        self.URL = self.URL.split("?")[0]
        self.URL = f"{self.URL}{self.URL_SUFFIX}"

    def acp(self, name="", acp_id=""):
        """
        Associate Access Control Policy with this HitCounts.

        Either the 'name' or the 'acp_id' is required, not both.
        :param name: (str) Name of ACP.
        :param acp_id: (str) UUID of ACP.
        """
        # either name or id of the ACP should be given
        logging.debug("In acp() for HitCounts class.")
        if acp_id != "":
            self.acp_id = acp_id
            self.URL = f"{self.fmc.configuration_url}{self.ACP_PREFIX_URL}/{self.acp_id}/operational/hitcounts"
        elif name != "":
            acp1 = AccessPolicies(fmc=self.fmc)
            acp1.get(name=name)
            if "id" in acp1.__dict__:
                self.URL = f"{self.fmc.configuration_url}{self.ACP_PREFIX_URL}/{acp1.id}/operational/hitcounts"
            else:
                logging.warning(
                    f'Access Control Policy "{name}" not found.  Cannot configure acp for HitCounts.'
                )
        else:
            logging.error("No accessPolicy name or id was provided.")

        # Rebuild the URL with possible new information
        self.URL = self.URL.split("?")[0]
        self.URL = f"{self.URL}{self.URL_SUFFIX}"

    def acp_rules(self, action=None, acp_rule_id=None, name=None):
        """
        Associate Access Rules to this HitCounts.

        Either the ID or the Name is required, not both.
        :param action: (str) 'add', 'remove', or 'clear'
        :param acp_rule_id: (str) UUID for Access Rule.
        :param name: (str) Name of Access Rule.
        """
        logging.debug("In acp_rules() for HitCounts class.")
        if name:
            # Need the acp_id if searching for AccessRule by name.
            if not self.acp_id:
                logging.error(
                    "Set AccessPolicy (acp_id or acp_name) prior to referencing acp_rules()."
                )
                return
            acp_rule = AccessRules(fmc=self.fmc, acp_id=self.acp_id)
            acp_rule.get(name=name)
            print(acp_rule.id)
            if hasattr(acp_rule, "id"):
                acp_rule_id = acp_rule.id
            else:
                acp_rule_id = None
                logging.warning(
                    f'AccessRule, "{name}", not found.  Cannot add to HitCounts.'
                )

        # If id is sent directly (and not looked up via name) assuming that the id is a valid AccessRule ID
        if acp_rule_id:
            if action == "add":
                if acp_rule_id not in self.acp_rule_ids:
                    self.acp_rule_ids.append(acp_rule_id)
                    logging.info(
                        f'Adding "{acp_rule_id}" to acp_rule_ids for this HitCounts.'
                    )
            elif action == "remove":
                if acp_rule_id in self.acp_rule_ids:
                    self.acp_rule_ids.remove(acp_rule_id)
                    logging.info(
                        f'Removed "{acp_rule_id}" from acp_rule_ids for this HitCounts.'
                    )
            elif action == "clear":
                self.acp_rule_ids = None
                logging.info("All ids removed from acp_rule_ids.")

        else:
            logging.warning(f"AccessRule ID not found.  Cannot add to HitCounts.")

        # Rebuild the URL with possible new information
        self.URL = self.URL.split("?")[0]
        self.URL = f"{self.URL}{self.URL_SUFFIX}"

    def prefilter_policy(self, name="", prefilter_policy_id=""):
        """
        Associate Pre-filter Policy with HitCount.

        Either the Name or the ID is needed, not both.
        :param name: (str) Name of PreFilter Policy.
        :param prefilter_policy_id: (str) UUID of PreFilter Policy.
        """
        # either name or id of the Prefilter Policy should be given
        logging.debug("In prefilter_policy() for HitCounts class.")
        if prefilter_policy_id != "":
            self.prefilter_id = prefilter_policy_id
            self.URL = (
                f"{self.fmc.configuration_url}{self.PREFILTER_PREFIX_URL}/"
                f"{self.prefilter_id}/operational/hitcounts"
            )
        elif name != "":
            ppolicy1 = PreFilterPolicies(fmc=self.fmc)
            ppolicy1.get(name=name)
            if "id" in ppolicy1.__dict__:
                self.prefilter_id = ppolicy1.id
                self.URL = (
                    f"{self.fmc.configuration_url}{self.PREFILTER_PREFIX_URL}/"
                    f"{self.prefilter_id}/operational/hitcounts"
                )
            else:
                logging.warning(
                    f'Access Control Policy "{name}" not found.  Cannot configure acp for HitCounts.'
                )
        else:
            logging.error("No accessPolicy name or id was provided.")

        # Rebuild the URL with possible new information
        self.URL = self.URL.split("?")[0]
        self.URL = f"{self.URL}{self.URL_SUFFIX}"

    def get(self, **kwargs):
        """
        Get HitCounts based on filter criteria.

        :return:
        """
        logging.debug("In get() for HitCount class.")
        self.parse_kwargs(**kwargs)

        if self.fmc.serverVersion < self.FIRST_SUPPORTED_FMC_VERSION:
            logging.error(
                f"Your FMC version, {self.fmc.serverVersion} does not support GET of this feature."
            )
            return {"items": []}
        if self.valid_for_get():
            if self.dry_run:
                logging.info(
                    "Dry Run enabled.  Not actually sending to FMC.  Here is what would have been sent:"
                )
                logging.info("\tMethod = GET")
                logging.info(f"\tURL = {self.URL}")
                return False
            response = self.fmc.send_to_api(method="get", url=self.URL)
            if "items" not in response:
                response["items"] = []
            return response
        else:
            logging.warning(
                "get() method failed due to failure to pass valid_for_get() test."
            )
            return False

    def put(self, **kwargs):
        """Though supported by FMC, API PUT method is not yet working for HitCounts in fmcapi."""
        logging.info(
            "Though supported by FMC, API PUT method is not yet working for HitCounts in fmcapi."
        )
        pass

    def delete(self, **kwargs):
        """Though supported by FMC, API DELETE method is not yet working for HitCounts in fmcapi."""
        logging.info(
            "Though supported by FMC, API DELETE method is not yet working for HitCounts in fmcapi."
        )
        pass

    def post(self):
        """POST method for HitCounts not supported."""
        logging.info("POST method for HitCounts not supported.")
        pass



"""Access Rules Classes."""

from fmcapi.api_objects.apiclasstemplate import APIClassTemplate
from fmcapi.api_objects.policy_services.accesspolicies import AccessPolicies
from fmcapi.api_objects.policy_services.intrusionpolicies import IntrusionPolicies
from fmcapi.api_objects.object_services.variablesets import VariableSets
from fmcapi.api_objects.object_services.securityzones import SecurityZones
from fmcapi.api_objects.object_services.vlantags import VlanTags
from fmcapi.api_objects.object_services.portobjectgroups import PortObjectGroups
from fmcapi.api_objects.object_services.protocolportobjects import ProtocolPortObjects
from fmcapi.api_objects.object_services.icmpv4objects import ICMPv4Objects
from fmcapi.api_objects.object_services.fqdns import FQDNS
from fmcapi.api_objects.object_services.networkgroups import NetworkGroups
from fmcapi.api_objects.object_services.networkaddresses import NetworkAddresses
from fmcapi.api_objects.policy_services.filepolicies import FilePolicies
from fmcapi.api_objects.object_services.isesecuritygrouptags import ISESecurityGroupTags
from fmcapi.api_objects.helper_functions import (
    get_networkaddress_type,
    true_false_checker,
)
from fmcapi.api_objects.object_services.applications import Applications
from fmcapi.api_objects.object_services.applicationfilters import ApplicationFilters
from fmcapi.api_objects.object_services.urlgroups import URLGroups
from fmcapi.api_objects.object_services.urls import URLs
from fmcapi.api_objects.object_services.realmusergroups import RealmUserGroups
from fmcapi.api_objects.object_services.realmusers import RealmUsers
import logging
import sys


class AccessRules(APIClassTemplate):
    """
    The AccessRules Object in the FMC.
    """

    VALID_JSON_DATA = [
        "id",
        "name",
        "type",
        "action",
        "enabled",
        "sendEventsToFMC",
        "logFiles",
        "logBegin",
        "logEnd",
        "variableSet",
        "originalSourceNetworks",
        "vlanTags",
        "users",
        "sourceNetworks",
        "destinationNetworks",
        "sourcePorts",
        "destinationPorts",
        "ipsPolicy",
        "urls",
        "sourceZones",
        "destinationZones",
        "applications",
        "filePolicy",
        "sourceSecurityGroupTags",
        "destinationSecurityGroupTags",
        "enableSyslog",
        "newComments",
        "commentHistoryList",
        "sourceDynamicObjects",
        "destinationDynamicObjects",
    ]
    VALID_FOR_KWARGS = VALID_JSON_DATA + [
        "acp_id",
        "acp_name",
        "insertBefore",
        "insertAfter",
        "section",
        "category",
    ]
    PREFIX_URL = "/policy/accesspolicies"
    REQUIRED_FOR_POST = ["name", "acp_id"]
    REQUIRED_FOR_GET = ["acp_id"]
    VALID_FOR_ACTION = [
        "ALLOW",
        "TRUST",
        "BLOCK",
        "MONITOR",
        "BLOCK_RESET",
        "BLOCK_INTERACTIVE",
        "BLOCK_RESET_INTERACTIVE",
    ]
    VALID_CHARACTERS_FOR_NAME = """[.\w\d_\-<>, ]"""

    @property
    def URL_SUFFIX(self):
        """
        Add the URL suffixes for categories, insertBefore and insertAfter
        NOTE: You must specify these at the time the object is initialized (created) for this feature
        to work correctly. Example:
            This works:
                new_rule = AccessRules(fmc=fmc, acp_name='acp1', insertBefore=2)

            This does not:
                new_rule = AccessRules(fmc=fmc, acp_name='acp1')
                new_rule.insertBefore = 2
        """
        url = "?"

        if "category" in self.__dict__:
            url = f"{url}category={self.category}&"
        if "insertBefore" in self.__dict__:
            url = f"{url}insertBefore={self.insertBefore}&"
        if "insertAfter" in self.__dict__:
            url = f"{url}insertAfter={self.insertAfter}&"
        if "insertBefore" in self.__dict__ and "insertAfter" in self.__dict__:
            logging.warning("ACP rule has both insertBefore and insertAfter params")
        if "section" in self.__dict__:
            url = f"{url}section={self.section}&"

        return url[:-1]

    @property
    def enabled(self):
        return self._enabled

    @enabled.setter
    def enabled(self, value=False):
        self._enabled = true_false_checker(value)

    @property
    def logBegin(self):
        return self._logBegin

    @logBegin.setter
    def logBegin(self, value=False):
        self._logBegin = true_false_checker(value)

    @property
    def logEnd(self):
        return self._logEnd

    @logEnd.setter
    def logEnd(self, value=False):
        self._logEnd = true_false_checker(value)

    @property
    def sendEventsToFMC(self):
        return self._sendEventsToFMC

    @sendEventsToFMC.setter
    def sendEventsToFMC(self, value=False):
        self._sendEventsToFMC = true_false_checker(value)

    @property
    def enableSyslog(self):
        return self._enableSyslog

    @enableSyslog.setter
    def enableSyslog(self, value=False):
        self._enableSyslog = true_false_checker(value)

    @property
    def newComments(self):
        return self._newComments

    @property
    def commentHistoryList(self):
        return self._commentHistoryList

    def __init__(self, fmc, **kwargs):
        """
        Initialize AccessRules object.

        Set self.type to "AccessRule", parse the kwargs, and set up the self.URL.

        :param fmc: (object) FMC object
        :param kwargs: Any other values passed during instantiation.
        :return: None
        """
        super().__init__(fmc, **kwargs)
        logging.debug("In __init__() for AccessRules class.")
        self.type = "AccessRule"
        self._enabled = False
        self._logBegin = False
        self._logEnd = False
        self._sendEventsToFMC = False
        self._enableSyslog = False
        self._newComments = []
        self._commentHistoryList = []
        self.parse_kwargs(**kwargs)
        self.URL = f"{self.URL}{self.URL_SUFFIX}"

    def format_data(self, filter_query=""):
        """
        Gather all the data in preparation for sending to API in JSON format.

        :param filter_query: (str) 'all' or 'kwargs'
        :return: (dict) json_data
        """
        json_data = super().format_data(filter_query=filter_query)
        logging.debug("In format_data() for AccessRules class.")
        if "sourceNetworks" in self.__dict__:
            json_data["sourceNetworks"] = {"objects": self.sourceNetworks["objects"]}
            json_data["sourceNetworks"]["literals"] = [
                {"type": v, "value": k}
                for k, v in self.sourceNetworks["literals"].items()
            ]
        if "destinationNetworks" in self.__dict__:
            json_data["destinationNetworks"] = {
                "objects": self.destinationNetworks["objects"]
            }
            json_data["destinationNetworks"]["literals"] = [
                {"type": v, "value": k}
                for k, v in self.destinationNetworks["literals"].items()
            ]

        if "action" in self.__dict__:
            if self.action not in self.VALID_FOR_ACTION:
                logging.warning(f"Action {self.action} is not a valid action.")
                logging.warning(f"\tValid actions are: {self.VALID_FOR_ACTION}.")
        return json_data

    def parse_kwargs(self, **kwargs):
        """
        Parse the kwargs and set self variables to match.

        :return: None
        """
        super().parse_kwargs(**kwargs)
        logging.debug("In parse_kwargs() for AccessRules class.")
        if "acp_id" in kwargs:
            self.acp(id=kwargs["acp_id"])
        if "acp_name" in kwargs:
            self.acp(name=kwargs["acp_name"])
        if "action" in kwargs:
            if kwargs["action"] in self.VALID_FOR_ACTION:
                self.action = kwargs["action"]
            else:
                logging.warning(f"Action {kwargs['action']} is not a valid action.")
                logging.warning(f"\tValid actions are: {self.VALID_FOR_ACTION}.")
        if "sourceNetworks" in kwargs:
            self.sourceNetworks = {"objects": [], "literals": {}}
            if kwargs["sourceNetworks"].get("objects"):
                self.sourceNetworks["objects"] = kwargs["sourceNetworks"]["objects"]
            if kwargs["sourceNetworks"].get("literals"):
                for literal in kwargs["sourceNetworks"]["literals"]:
                    self.sourceNetworks["literals"][literal["value"]] = literal["type"]
        if "destinationNetworks" in kwargs:
            self.destinationNetworks = {"objects": [], "literals": {}}
            if kwargs["destinationNetworks"].get("objects"):
                self.destinationNetworks["objects"] = kwargs["destinationNetworks"][
                    "objects"
                ]
            if kwargs["destinationNetworks"].get("literals"):
                for literal in kwargs["destinationNetworks"]["literals"]:
                    self.destinationNetworks["literals"][literal["value"]] = literal[
                        "type"
                    ]
        if "enableSyslog" in kwargs:
            self.enableSyslog = kwargs["enableSyslog"]
        if "logBegin" in kwargs:
            self.logBegin = kwargs["logBegin"]
        if "logEnd" in kwargs:
            self.logEnd = kwargs["logEnd"]
        if "enabled" in kwargs:
            self.enabled = kwargs["enabled"]
        if "sendEventsToFMC" in kwargs:
            self.sendEventsToFMC = kwargs["sendEventsToFMC"]
        if "newComments" in kwargs:
            self._newComments = kwargs["newComments"]
        if "commentHistoryList" in kwargs:
            self._commentHistoryList = kwargs["commentHistoryList"]
        if "users" in kwargs:
            self.users = {"objects": []}
            if kwargs["users"].get("objects"):
                self.users["objects"] = kwargs["users"]["objects"]

    def acp(self, name="", id=""):
        """
        Associate an AccessPolicies object with this AccessRule object.

        :param name: (str)  Name of ACP.
        :param id: (str) ID of ACP.
        :return: None
        """
        # either name or id of the ACP should be given
        logging.debug("In acp() for AccessRules class.")
        if id != "":
            self.acp_id = id
            self.URL = f"{self.fmc.configuration_url}{self.PREFIX_URL}/{id}/accessrules"
        elif name != "":
            acp1 = AccessPolicies(fmc=self.fmc)
            acp1.get(name=name)
            if "id" in acp1.__dict__:
                self.acp_id = acp1.id
                self.URL = f"{self.fmc.configuration_url}{self.PREFIX_URL}/{acp1.id}/accessrules"
            else:
                logging.warning(
                    f"Access Control Policy {name} not found.  Cannot set up accessPolicy for AccessRules."
                )
        else:
            logging.error("No accessPolicy name or ID was provided.")

    def intrusion_policy(self, action, name=""):
        """
        Add/remove name of ipsPolicy field of AccessRules object.

        :param action: (str) 'set', or 'clear'
        :param name: (str) Name of intrusion policy in FMC.
        :return: None
        """
        logging.debug("In intrusion_policy() for AccessRules class.")
        if action == "clear":
            if "ipsPolicy" in self.__dict__:
                del self.ipsPolicy
                logging.info("Intrusion Policy removed from this AccessRules object.")
        elif action == "set":
            ips = IntrusionPolicies(fmc=self.fmc, name=name)
            ips.get()
            self.ipsPolicy = {"name": ips.name, "id": ips.id, "type": ips.type}
            logging.info(
                f'Intrusion Policy set to "{name}" for this AccessRules object.'
            )

    def variable_set(self, action, name="Default-Set"):
        """
        Add/remove name of variableSet field of AccessRules object.

        :param action: (str) 'set', or 'clear'
        :param name: (str) Name of variable set in FMC.
        :return: None
        """
        logging.debug("In variable_set() for AccessRules class.")
        if action == "clear":
            if "variableSet" in self.__dict__:
                del self.variableSet
                logging.info("Variable Set removed from this AccessRules object.")
        elif action == "set":
            vs = VariableSets(fmc=self.fmc)
            vs.get(name=name)
            self.variableSet = {"name": vs.name, "id": vs.id, "type": vs.type}
            logging.info(f'VariableSet set to "{name}" for this AccessRules object.')

    def file_policy(self, action, name="None"):
        """
        Add/remove name of filePolicy field of AccessRules object.

        :param action: (str) 'set', or 'clear'
        :param name: (str) Name of file policy in FMC.
        :return: None
        """
        logging.debug("In file_policy() for ACPRule class.")
        if action == "clear":
            if "filePolicy" in self.__dict__:
                del self.filePolicy
                logging.info("file_policy removed from this AccessRules object.")
        elif action == "set":
            fp = FilePolicies(fmc=self.fmc)
            fp.get(name=name)
            self.filePolicy = {"name": fp.name, "id": fp.id, "type": fp.type}
            logging.info(f'file_policy set to "{name}" for this AccessRules object.')

    def vlan_tags(self, action, name=""):
        """
        Add/modify name to vlanTags field of AccessRules object.

        :param action: (str) 'add', 'remove', or 'clear'
        :param name: (str) Name of VLAN tag in FMC.
        :return: None
        """
        logging.debug("In vlan_tags() for AccessRules class.")
        if action == "add":
            vlantag = VlanTags(fmc=self.fmc)
            vlantag.get(name=name)
            if "id" in vlantag.__dict__:
                if "vlanTags" in self.__dict__:
                    new_vlan = {
                        "name": vlantag.name,
                        "id": vlantag.id,
                        "type": vlantag.type,
                    }
                    duplicate = False
                    for obj in self.vlanTags["objects"]:
                        if obj["name"] == new_vlan["name"]:
                            duplicate = True
                            break
                    if not duplicate:
                        self.vlanTags["objects"].append(new_vlan)
                        logging.info(
                            f'Adding "{name}" to vlanTags for this AccessRules.'
                        )
                else:
                    self.vlanTags = {
                        "objects": [
                            {
                                "name": vlantag.name,
                                "id": vlantag.id,
                                "type": vlantag.type,
                            }
                        ]
                    }
                    logging.info(f'Adding "{name}" to vlanTags for this AccessRules.')
            else:
                logging.warning(
                    f'VLAN Tag, "{name}", not found.  Cannot add to AccessRules.'
                )
        elif action == "remove":
            vlantag = VlanTags(fmc=self.fmc)
            vlantag.get(name=name)
            if "id" in vlantag.__dict__:
                if "vlanTags" in self.__dict__:
                    objects = []
                    for obj in self.vlanTags["objects"]:
                        if obj["name"] != name:
                            objects.append(obj)
                    self.vlanTags["objects"] = objects
                    logging.info(
                        f'Removed "{name}" from vlanTags for this AccessRules.'
                    )
                else:
                    logging.info(
                        "vlanTags doesn't exist for this AccessRules.  Nothing to remove."
                    )
            else:
                logging.warning(
                    f"VLAN Tag, {name}, not found.  Cannot remove from AccessRules."
                )
        elif action == "clear":
            if "vlanTags" in self.__dict__:
                del self.vlanTags
                logging.info("All VLAN Tags removed from this AccessRules object.")

    def source_zone(self, action, name=""):
        """
        Add/modify name to sourceZones field of AccessRules object.

        :param action: (str) 'add', 'remove', or 'clear'
        :param name: (str) Name of Security Zone in FMC.
        :return: None
        """
        logging.debug("In source_zone() for AccessRules class.")
        if action == "add":
            sz = SecurityZones(fmc=self.fmc)
            sz.get(name=name)
            if "id" in sz.__dict__:
                if "sourceZones" in self.__dict__:
                    new_zone = {"name": sz.name, "id": sz.id, "type": sz.type}
                    duplicate = False
                    for obj in self.sourceZones["objects"]:
                        if obj["name"] == new_zone["name"]:
                            duplicate = True
                            break
                    if not duplicate:
                        self.sourceZones["objects"].append(new_zone)
                        logging.info(
                            f'Adding "{name}" to sourceZones for this AccessRules.'
                        )
                else:
                    self.sourceZones = {
                        "objects": [{"name": sz.name, "id": sz.id, "type": sz.type}]
                    }
                    logging.info(
                        f'Adding "{name}" to sourceZones for this AccessRules.'
                    )
            else:
                logging.warning(
                    'Security Zone, "{name}", not found.  Cannot add to AccessRules.'
                )
        elif action == "remove":
            sz = SecurityZones(fmc=self.fmc)
            sz.get(name=name)
            if "id" in sz.__dict__:
                if "sourceZones" in self.__dict__:
                    objects = []
                    for obj in self.sourceZones["objects"]:
                        if obj["name"] != name:
                            objects.append(obj)
                    self.sourceZones["objects"] = objects
                    logging.info(
                        f'Removed "{name}" from sourceZones for this AccessRules.'
                    )
                else:
                    logging.info(
                        "sourceZones doesn't exist for this AccessRules.  Nothing to remove."
                    )
            else:
                logging.warning(
                    f'Security Zone, "{name}", not found.  Cannot remove from AccessRules.'
                )
        elif action == "clear":
            if "sourceZones" in self.__dict__:
                del self.sourceZones
                logging.info("All Source Zones removed from this AccessRules object.")

    def destination_zone(self, action, name=""):
        """
        Add/modify name to destinationZones field of AccessRules object.

        :param action: (str) 'add', 'remove', or 'clear'
        :param name: (str) Name of Security Zone in FMC.
        :return: None
        """
        logging.debug("In destination_zone() for AccessRules class.")
        if action == "add":
            sz = SecurityZones(fmc=self.fmc)
            sz.get(name=name)
            if "id" in sz.__dict__:
                if "destinationZones" in self.__dict__:
                    new_zone = {"name": sz.name, "id": sz.id, "type": sz.type}
                    duplicate = False
                    for obj in self.destinationZones["objects"]:
                        if obj["name"] == new_zone["name"]:
                            duplicate = True
                            break
                    if not duplicate:
                        self.destinationZones["objects"].append(new_zone)
                        logging.info(
                            f'Adding "{name}" to destinationZones for this AccessRules.'
                        )
                else:
                    self.destinationZones = {
                        "objects": [{"name": sz.name, "id": sz.id, "type": sz.type}]
                    }
                    logging.info(
                        f'Adding "{name}" to destinationZones for this AccessRules.'
                    )
            else:
                logging.warning(
                    f'Security Zone, "{name}", not found.  Cannot add to AccessRules.'
                )
        elif action == "remove":
            sz = SecurityZones(fmc=self.fmc)
            sz.get(name=name)
            if "id" in sz.__dict__:
                if "destinationZones" in self.__dict__:
                    objects = []
                    for obj in self.destinationZones["objects"]:
                        if obj["name"] != name:
                            objects.append(obj)
                    self.destinationZones["objects"] = objects
                    logging.info(
                        'Removed "{name}" from destinationZones for this AccessRules.'
                    )
                else:
                    logging.info(
                        "destinationZones doesn't exist for this AccessRules.  Nothing to remove."
                    )
            else:
                logging.warning(
                    f"Security Zone, {name}, not found.  Cannot remove from AccessRules."
                )
        elif action == "clear":
            if "destinationZones" in self.__dict__:
                del self.destinationZones
                logging.info(
                    "All Destination Zones removed from this AccessRules object."
                )

    def source_port(self, action, name=""):
        """
        Add/modify name to sourcePorts field of AccessRules object.

        :param action: (str) 'add', 'addgroup', 'remove', or 'clear'
        :param name: (str) Name of Port in FMC.
        :return: None
        """
        logging.debug("In source_port() for AccessRules class.")
        if action == "add":
            pport_json = ProtocolPortObjects(fmc=self.fmc)
            pport_json.get(name=name)
            icmpv4_json = ICMPv4Objects(fmc=self.fmc)
            icmpv4_json.get(name=name)
            if "id" in pport_json.__dict__:
                item = pport_json
            elif "id" in icmpv4_json.__dict__:
                item = icmpv4_json
            else:
                item = PortObjectGroups(fmc=self.fmc)
                item.get(name=name)
            if "id" in item.__dict__:
                if "sourcePorts" in self.__dict__:
                    new_port = {"name": item.name, "id": item.id, "type": item.type}
                    duplicate = False
                    if "objects" not in self.sourcePorts:
                        self.__dict__["sourcePorts"]["objects"] = []
                    for obj in self.sourcePorts["objects"]:
                        if obj["name"] == new_port["name"]:
                            duplicate = True
                            break
                    if not duplicate:
                        self.sourcePorts["objects"].append(new_port)
                        logging.info(
                            f'Adding "{name}" to sourcePorts for this AccessRules.'
                        )
                else:
                    self.sourcePorts = {
                        "objects": [
                            {"name": item.name, "id": item.id, "type": item.type}
                        ]
                    }
                    logging.info(
                        f'Adding "{name}" to sourcePorts for this AccessRules.'
                    )
            else:
                logging.warning(
                    f'Protocol Port or Protocol Port Group: "{name}", '
                    f"not found.  Cannot add to AccessRules."
                )
        elif action == "addgroup":
            item = PortObjectGroups(fmc=self.fmc)
            item.get(name=name)
            if "id" in item.__dict__:
                if "sourcePorts" in self.__dict__:
                    new_port = {"name": item.name, "id": item.id, "type": item.type}
                    duplicate = False
                    if "objects" not in self.sourcePorts:
                        self.__dict__["sourcePorts"]["objects"] = []
                    for obj in self.sourcePorts["objects"]:
                        if obj["name"] == new_port["name"]:
                            duplicate = True
                            break
                    if not duplicate:
                        self.sourcePorts["objects"].append(new_port)
                        logging.info(
                            f'Adding "{name}" to sourcePorts for this AccessRules.'
                        )
                else:
                    self.sourcePorts = {
                        "objects": [
                            {"name": item.name, "id": item.id, "type": item.type}
                        ]
                    }
                    logging.info(
                        f'Adding "{name}" to sourcePorts for this AccessRules.'
                    )
            else:
                logging.warning(
                    f'Protocol Port Port Group: "{name}", '
                    f"not found.  Cannot add to AccessRules."
                )
        elif action == "remove":
            pport_json = ProtocolPortObjects(fmc=self.fmc)
            pport_json.get(name=name)
            icmpv4_json = ICMPv4Objects(fmc=self.fmc)
            icmpv4_json.get(name=name)
            if "id" in pport_json.__dict__:
                item = pport_json
            elif "id" in icmpv4_json.__dict__:
                item = icmpv4_json
            else:
                item = PortObjectGroups(fmc=self.fmc)
                item.get(name=name)
            if "id" in item.__dict__:
                if "sourcePorts" in self.__dict__:
                    objects = []
                    for obj in self.sourcePorts["objects"]:
                        if obj["name"] != name:
                            objects.append(obj)
                    self.sourcePorts["objects"] = objects
                    logging.info(
                        f'Removed "{name}" from sourcePorts for this AccessRules.'
                    )
                else:
                    logging.info(
                        "sourcePorts doesn't exist for this AccessRules.  Nothing to remove."
                    )
            else:
                logging.warning(
                    f'Protocol Port or Protocol Port Group: "{name}", '
                    f"not found.  Cannot add to AccessRules."
                )
        elif action == "clear":
            if "sourcePorts" in self.__dict__:
                del self.sourcePorts
                logging.info("All Source Ports removed from this AccessRules object.")

    def destination_port(self, action, name=""):
        """
        Add/modify name to destinationPorts field of AccessRules object.

        :param action: (str) 'add', 'addgroup', 'remove', or 'clear'
        :param name: (str) Name of Port in FMC.
        :return: None
        """
        logging.debug("In destination_port() for AccessRules class.")
        if action == "add":
            pport_json = ProtocolPortObjects(fmc=self.fmc)
            pport_json.get(name=name)
            icmpv4_json = ICMPv4Objects(fmc=self.fmc)
            icmpv4_json.get(name=name)
            if "id" in pport_json.__dict__:
                item = pport_json
            elif "id" in icmpv4_json.__dict__:
                item = icmpv4_json
            else:
                item = PortObjectGroups(fmc=self.fmc)
                item.get(name=name)
            if "id" in item.__dict__:
                if "destinationPorts" in self.__dict__:
                    new_port = {"name": item.name, "id": item.id, "type": item.type}
                    duplicate = False
                    if "objects" not in self.destinationPorts:
                        self.__dict__["destinationPorts"]["objects"] = []
                    for obj in self.destinationPorts["objects"]:
                        if obj["name"] == new_port["name"]:
                            duplicate = True
                            break
                    if not duplicate:
                        self.destinationPorts["objects"].append(new_port)
                        logging.info(
                            f'Adding "{name}" to destinationPorts for this AccessRules.'
                        )
                else:
                    self.destinationPorts = {
                        "objects": [
                            {"name": item.name, "id": item.id, "type": item.type}
                        ]
                    }
                    logging.info(
                        f'Adding "{name}" to destinationPorts for this AccessRules.'
                    )
            else:
                logging.warning(
                    f'Protocol Port or Protocol Port Group: "{name}", '
                    f"not found.  Cannot add to AccessRules."
                )
        if action == "addgroup":
            item = PortObjectGroups(fmc=self.fmc)
            item.get(name=name)
            if "id" in item.__dict__:
                if "destinationPorts" in self.__dict__:
                    new_port = {"name": item.name, "id": item.id, "type": item.type}
                    duplicate = False
                    if "objects" not in self.destinationPorts:
                        self.__dict__["destinationPorts"]["objects"] = []
                    for obj in self.destinationPorts["objects"]:
                        if obj["name"] == new_port["name"]:
                            duplicate = True
                            break
                    if not duplicate:
                        self.destinationPorts["objects"].append(new_port)
                        logging.info(
                            f'Adding "{name}" to destinationPorts for this AccessRules.'
                        )
                else:
                    self.destinationPorts = {
                        "objects": [
                            {"name": item.name, "id": item.id, "type": item.type}
                        ]
                    }
                    logging.info(
                        f'Adding "{name}" to destinationPorts for this AccessRules.'
                    )
            else:
                logging.warning(
                    f'Protocol Port Port Group: "{name}", '
                    f"not found.  Cannot add to AccessRules."
                )
        elif action == "remove":
            pport_json = ProtocolPortObjects(fmc=self.fmc)
            pport_json.get(name=name)
            icmpv4_json = ICMPv4Objects(fmc=self.fmc)
            icmpv4_json.get(name=name)
            if "id" in pport_json.__dict__:
                item = pport_json
            elif "id" in icmpv4_json.__dict__:
                item = icmpv4_json
            else:
                item = PortObjectGroups(fmc=self.fmc)
                item.get(name=name)
            if "id" in item.__dict__:
                if "destinationPorts" in self.__dict__:
                    objects = []
                    for obj in self.destinationPorts["objects"]:
                        if obj["name"] != name:
                            objects.append(obj)
                    self.destinationPorts["objects"] = objects
                    logging.info(
                        f'Removed "{name}" from destinationPorts for this AccessRules.'
                    )
                else:
                    logging.info(
                        "destinationPorts doesn't exist for this AccessRules.  Nothing to remove."
                    )
            else:
                logging.warning(
                    f'Protocol Port or Protocol Port Group: "{name}", '
                    f"not found.  Cannot add to AccessRules."
                )
        elif action == "clear":
            if "destinationPorts" in self.__dict__:
                del self.destinationPorts
                logging.info(
                    "All Destination Ports removed from this AccessRules object."
                )

    def source_network(self, action, name="", literal=None):
        """
        Add/modify name/literal to sourceNetworks field of AccessRules object.

        :param action: (str) the action to be done 'add', 'remove', 'clear'
        :param name: (str) name of the object in question
        :param literal: (dict) the literal in question {value:<>, type:<>}
        :return: None
        """
        # using dict() as default value is dangerous here, any thoughts/workarounds on this?
        logging.debug("In source_network() for AccessRules class.")
        if literal and name != "":
            raise ValueError(
                "Only one of literals or name (object name) should be set while creating a source network"
            )

        if not hasattr(self, "sourceNetworks"):
            self.sourceNetworks = {"objects": [], "literals": {}}

        if action == "add":
            if literal:
                type_ = get_networkaddress_type(literal)
                self.sourceNetworks["literals"][literal] = type_
                logging.info(
                    f'Adding literal "{literal}" of type "{type_}" to sourceNetworks for this AccessRules.'
                )
            else:
                ipaddresses_json = NetworkAddresses(fmc=self.fmc).get()
                networkgroup_json = NetworkGroups(fmc=self.fmc).get()
                fqdns_json = FQDNS(fmc=self.fmc).get()
                items = (
                    ipaddresses_json.get("items", [])
                    + networkgroup_json.get("items", [])
                    + fqdns_json.get("items", [])
                )
                new_net = None
                for item in items:
                    if item["name"] == name:
                        new_net = {
                            "name": item["name"],
                            "id": item["id"],
                            "type": item["type"],
                        }
                        break
                if new_net is None:
                    logging.warning(
                        f'Network "{name}" is not found in FMC.  Cannot add to sourceNetworks.'
                    )
                else:
                    if "sourceNetworks" in self.__dict__:
                        # thus either some objects are already present in sourceNetworks,
                        # or only literals are present in sourceNetworks
                        if "objects" in self.__dict__["sourceNetworks"]:
                            # some objects are already present
                            duplicate = False
                            # see if its a duplicate or not. If not, append to the list of
                            # existing objects in sourceNetworks
                            for obj in self.sourceNetworks["objects"]:
                                if obj["name"] == new_net["name"]:
                                    duplicate = True
                                    break
                            if not duplicate:
                                self.sourceNetworks["objects"].append(new_net)
                                logging.info(
                                    f'Adding "{name}" to sourceNetworks for this AccessRules.'
                                )
                        else:
                            # this means no objects were present in sourceNetworks,
                            # and sourceNetworks contains literals only
                            self.sourceNetworks.update({"objects": [new_net]})
                            # So update the sourceNetworks dict which contained 'literals' key initially
                            # to have a 'objects' key as well
                            logging.info(
                                f'Adding "{name}" to sourceNetworks for this AccessRules.'
                            )
                    else:
                        # None of literals or objects are present in sourceNetworks,
                        # so initialize it with objects and update the provided object
                        self.sourceNetworks = {"objects": [new_net]}
                        logging.info(
                            f'Adding "{name}" to sourceNetworks for this AccessRules.'
                        )
        elif action == "remove":
            if "sourceNetworks" in self.__dict__:
                if name != "":
                    # an object's name has been provided to be removed
                    objects = []
                    for obj in self.sourceNetworks["objects"]:
                        if obj["name"] != name:
                            objects.append(obj)
                    if len(objects) == 0:
                        # it was the last object which was deleted now
                        del self.sourceNetworks
                        logging.info(
                            f'Removed "{name}" from sourceNetworks for this AccessRules'
                        )
                        logging.info(
                            "All Source Networks removed from this AccessRules object."
                        )
                    else:
                        self.sourceNetworks["objects"] = objects
                        logging.info(
                            f'Removed "{name}" from sourceNetworks for this AccessRules.'
                        )
                else:
                    # a literal value has been provided to be removed
                    type_ = self.sourceNetworks["literals"].get(literal)
                    if type_:
                        self.sourceNetworks["literals"].pop(literal)
                        logging.info(
                            f'Removed literal "{literal}" of type '
                            f'"{type_}" from sourceNetworks for this AccessRules.'
                        )
                    else:
                        logging.info(
                            f'Unable to removed literal "{literal}" from sourceNetworks as it was not found'
                        )
            else:
                logging.info(
                    "sourceNetworks doesn't exist for this AccessRules.  Nothing to remove."
                )
        elif action == "clear":
            if "sourceNetworks" in self.__dict__:
                del self.sourceNetworks
                logging.info(
                    "All Source Networks removed from this AccessRules object."
                )

    def destination_network(self, action, name="", literal=None):
        """
        Add/modify name/literal to destinationNetworks field of AccessRules object.

        :param action: (str) the action to be done 'add', 'remove', 'clear'
        :param name: (str) name of the object in question
        :param literal: (dict) the literal in question {value:<>, type:<>}
        :return: None
        """
        # using dict() as default value is dangerous here, any thoughts/workarounds on this?

        logging.debug("In destination_network() for ACPRule class.")
        if literal and name != "":
            raise ValueError(
                "Only one of literals or name (object name) should be set while creating a source network"
            )

        if not hasattr(self, "destinationNetworks"):
            self.destinationNetworks = {"objects": [], "literals": {}}

        if action == "add":
            if literal:
                type_ = get_networkaddress_type(literal)
                self.destinationNetworks["literals"][literal] = type_
                logging.info(
                    f'Adding literal "{literal}" of type "{type_}" '
                    f"to destinationNetworks for this AccessRules."
                )
            else:
                ipaddresses_json = NetworkAddresses(fmc=self.fmc).get()
                networkgroup_json = NetworkGroups(fmc=self.fmc).get()
                if self.fmc.serverVersion >= "6.4":
                    fqdns_json = FQDNS(fmc=self.fmc).get()
                else:
                    fqdns_json = {"items": []}
                items = (
                    ipaddresses_json.get("items", [])
                    + networkgroup_json.get("items", [])
                    + fqdns_json.get("items", [])
                )
                new_net = None
                for item in items:
                    if item["name"] == name:
                        new_net = {
                            "name": item["name"],
                            "id": item["id"],
                            "type": item["type"],
                        }
                        break
                if new_net is None:
                    logging.warning(
                        f'Network "{name}" is not found in FMC.  Cannot add to destinationNetworks.'
                    )
                else:
                    if "destinationNetworks" in self.__dict__:
                        # thus either some objects are already present in destinationNetworks,
                        # or only literals are present in destinationNetworks
                        if "objects" in self.__dict__["destinationNetworks"]:
                            # some objects are already present
                            duplicate = False
                            for obj in self.destinationNetworks["objects"]:
                                if obj["name"] == new_net["name"]:
                                    duplicate = True
                                    break
                            if not duplicate:
                                self.destinationNetworks["objects"].append(new_net)
                                logging.info(
                                    f'Adding "{name}" to destinationNetworks for this AccessRules.'
                                )
                        else:
                            # this means no objects were present in destinationNetworks,
                            # and destinationNetworks contains literals only
                            self.destinationNetworks.update({"objects": [new_net]})
                            # So update the destinationNetworks dict which contained 'literals' key initially
                            # to have a 'objects' key as well
                            logging.info(
                                f'Adding "{name}" to destinationNetworks for this AccessRules.'
                            )
                    else:
                        # None of literals or objects are present in destinationNetworks,
                        # so initialize it with objects and update the provided object
                        self.destinationNetworks = {"objects": [new_net]}
                        logging.info(
                            f'Adding "{name}" to destinationNetworks for this AccessRules.'
                        )
        elif action == "remove":
            if "destinationNetworks" in self.__dict__:
                if name != "":
                    # an object's name has been provided to be removed
                    objects = []
                    for obj in self.destinationNetworks["objects"]:
                        if obj["name"] != name:
                            objects.append(obj)
                    if len(objects) == 0:
                        # it was the last object which was deleted now
                        del self.destinationNetworks
                        logging.info(
                            f'Removed "{name}" from destinationNetworks for this AccessRules'
                        )
                        logging.info(
                            "All Destination Networks removed from this AccessRules object."
                        )
                    else:
                        self.destinationNetworks["objects"] = objects
                        logging.info(
                            f'Removed "{name}" from destinationNetworks for this AccessRules.'
                        )
                else:
                    # a literal value has been provided to be removed
                    type_ = self.destinationNetworks["literals"].get(literal)
                    if type_:
                        self.destinationNetworks["literals"].pop(literal)
                        logging.info(
                            f'Removed literal "{literal}" of '
                            f'type "{type_}" from destinationNetworks for this AccessRules.'
                        )
                    else:
                        logging.info(
                            f'Unable to removed literal "{literal}" '
                            f"from destinationNetworks as it was not found"
                        )
            else:
                logging.info(
                    "destinationNetworks doesn't exist for this AccessRules.  Nothing to remove."
                )
        elif action == "clear":
            if "destinationNetworks" in self.__dict__:
                del self.destinationNetworks
                logging.info(
                    "All Destination Networks removed from this AccessRules object."
                )

    def source_sgt(self, action, name="", literal=None):
        """
        Add/modify name/literal to the sourceSecurityGroupTags field of AccessRules object.

        :param action: (str) 'add', 'remove', or 'clear'
        :param name: (str) Name of SGT in FMC.
        :param literal: (dict) {value:<>, type:<>}
        :return: None
        """
        # using dict() as default value is dangerous here, any thoughts/workarounds on this?

        logging.debug("In source_sgt() for ACPRule class.")
        if literal and name != "":
            raise ValueError(
                "Only one of literals or name (object name) should be set while creating a source sgt"
            )

        if not hasattr(self, "sourceSecurityGroupTags"):
            self.sourceSecurityGroupTags = {"objects": [], "literals": {}}

        if action == "add":
            if literal:
                type_ = "ISESecurityGroupTag"
                self.sourceSecurityGroupTags["literals"][literal] = type_
                logging.info(
                    f'Adding literal "{literal}" of type "{type_}" '
                    f"to sourceSecurityGroupTags for this AccessRules."
                )
            else:
                # Query FMC for all SGTs and iterate through them to see if our name matches 1 of them.
                sgt = ISESecurityGroupTags(fmc=self.fmc)
                sgt.get(name=name)
                if "id" in sgt.__dict__:
                    item = sgt
                else:
                    item = {}
                new_sgt = None
                if item.name == name:
                    new_sgt = {"name": item.name, "tag": item.tag, "type": item.type}
                if new_sgt is None:
                    logging.warning(
                        f'SecurityGroupTag "{name}" is not found in FMC.  '
                        f"Cannot add to sourceSecurityGroupTags."
                    )
                else:
                    if "sourceSecurityGroupTags" in self.__dict__:
                        # thus either some objects are already present in sourceSecurityGroupTags,
                        # or only literals are present in sourceSecurityGroupTags
                        if "objects" in self.__dict__["sourceSecurityGroupTags"]:
                            # some objects are already present
                            duplicate = False
                            for obj in self.sourceSecurityGroupTags["objects"]:
                                if obj["name"] == new_sgt["name"]:
                                    duplicate = True
                                    break
                            if not duplicate:
                                self.sourceSecurityGroupTags["objects"].append(new_sgt)
                                logging.info(
                                    f'Adding "{name}" to sourceSecurityGroupTags for this AccessRules.'
                                )
                        else:
                            # this means no objects were present in sourceSecurityGroupTags,
                            # and sourceSecurityGroupTags contains literals only
                            self.sourceSecurityGroupTags.update({"objects": [new_sgt]})
                            # So update the sourceSecurityGroupTags dict which contained 'literals' key initially
                            # to have a 'objects' key as well
                            logging.info(
                                f'Adding "{name}" to sourceSecurityGroupTags for this AccessRules.'
                            )
                    else:
                        # None of literals or objects are present in sourceSecurityGroupTags,
                        # so initialize it with objects and update the provided object
                        self.sourceSecurityGroupTags = {"objects": [new_sgt]}
                        logging.info(
                            f'Adding "{name}" to sourceSecurityGroupTags for this AccessRules.'
                        )
        elif action == "remove":
            if "sourceSecurityGroupTags" in self.__dict__:
                if name != "":
                    # an object's name has been provided to be removed
                    objects = []
                    for obj in self.sourceSecurityGroupTags["objects"]:
                        if obj["name"] != name:
                            objects.append(obj)
                    if len(objects) == 0:
                        # it was the last object which was deleted now
                        del self.sourceSecurityGroupTags
                        logging.info(
                            f'Removed "{name}" from sourceSecurityGroupTags for this AccessRules'
                        )
                        logging.info(
                            "All source security group tags are removed from this AccessRules object."
                        )
                    else:
                        self.sourceSecurityGroupTags["objects"] = objects
                        logging.info(
                            f'Removed "{name}" from sourceSecurityGroupTags for this AccessRules.'
                        )
                else:
                    # a literal value has been provided to be removed
                    type_ = self.sourceSecurityGroupTags["literals"].get(literal)
                    if type_:
                        self.sourceSecurityGroupTags["literals"].pop(literal)
                        logging.info(
                            f'Removed literal "{literal}" of '
                            f'type "{type_}" from sourceSecurityGroupTags for this AccessRules.'
                        )
                    else:
                        logging.info(
                            f'Unable to removed literal "{literal}" '
                            f"from sourceSecurityGroupTags as it was not found"
                        )
            else:
                logging.info(
                    "No sourceSecurityGroupTags exist for this AccessRules.  Nothing to remove."
                )
        elif action == "clear":
            if "sourceSecurityGroupTags" in self.__dict__:
                del self.sourceSecurityGroupTags
                logging.info(
                    "All source security group tags are removed from this AccessRules object."
                )

    def destination_sgt(self, action, name="", literal=None):
        """
        Add/modify name/literal to the destinationSecurityGroupTags field of AccessRules object.

        :param action: (str) 'add', 'remove', or 'clear'
        :param name: (str) Name of SGT in FMC.
        :param literal: (dict) {value:<>, type:<>}
        :return: None
        """
        pass

    def realm_user(self, action, name="", realm_type="user"):
        """
        Add/modify realm users field of AccessRules object.
        :param action: (str) 'add', 'remove', or 'clear'
        :param name: (str) Name Realm user/usergroup in FMC.
        :param realm_type (str) 'user' or 'group' default to user - not used in 'clear'
        :return: None
        NOTE Only tested on single realm deployment
        """
        logging.debug("In realm_user() for AccessRules class.")
        if action == "add":
            if realm_type == "user":
                user = RealmUsers(fmc=self.fmc)
                logging.debug("Type user")
            elif realm_type == "group":
                user = RealmUserGroups(fmc=self.fmc)
                logging.debug("Type Group")
            else:
                logging.warning("not a valid realm_type - 'user'|'group'")
                return None
            user.get(name=name)
            if "id" in user.__dict__:
                if "users" in self.__dict__:
                    new_realm_user = {
                        "name": user.name,
                        "id": user.id,
                        "type": user.type,
                        "realm": user.realm,
                    }
                    duplicate = False
                    if "objects" not in self.users:
                        self.__dict__["users"]["objects"] = []

                    for obj in self.users["objects"]:
                        if obj["name"] == new_realm_user["name"]:
                            duplicate = True
                            break
                    if not duplicate:
                        self.users["objects"].append(new_realm_user)
                        logging.info(f'Adding "{name}" to users for this AccessRules.')
                else:
                    self.users = {
                        "objects": [
                            {
                                "name": user.name,
                                "id": user.id,
                                "type": user.type,
                                "realm": user.realm,
                            }
                        ]
                    }
                    logging.info(f'Adding "{name}" to users for this AccessRules.')
            else:
                logging.warning(
                    f'User: "{name}", ' f"not found.  Cannot add to AccessRules."
                )

        elif action == "remove":
            if realm_type == "user":
                user = RealmUsers(fmc=self.fmc)
            elif realm_type == "group":
                user = RealmUserGroups(fmc=self.fmc)
            user.get(name=name)
            if "id" in user.__dict__:
                if "users" in self.__dict__:
                    users = []
                    for obj in self.users["objects"]:
                        if obj["name"] != name:
                            users.append(obj)
                    self.users["objects"] = users
                    logging.info(f'Removed "{name}" from users for this AccessRules.')
                else:
                    logging.info(
                        "Users doesn't exist for this AccessRules.  Nothing to remove."
                    )
            else:
                logging.warning(
                    f"User, {name}, not found.  Cannot remove from AccessRules."
                )

        elif action == "clear":
            if "users" in self.__dict__:
                del self.users
                logging.info("All Users removed from this AccessRules object.")

    def application(self, action, name=""):
        """
        Add/modify name to applications field of AccessRules object.
        :param action: (str) 'add', 'remove', or 'clear'
        :param name: (str) Name of Application in FMC.
        :return: None
        """
        logging.debug("In application() for AccessRules class.")
        if action == "add":
            app = Applications(fmc=self.fmc)
            app.get(name=name)
            if "id" in app.__dict__:
                if "applications" in self.__dict__:
                    new_app = {"name": app.name, "id": app.id, "type": app.type}
                    duplicate = False
                    if "applications" not in self.applications:
                        self.__dict__["applications"]["applications"] = []
                    for obj in self.applications["applications"]:
                        if obj["name"] == new_app["name"]:
                            duplicate = True
                            break
                    if not duplicate:
                        self.applications["applications"].append(new_app)
                        logging.info(
                            f'Adding "{name}" to applications for this AccessRules.'
                        )
                else:
                    self.applications = {
                        "applications": [
                            {"name": app.name, "id": app.id, "type": app.type}
                        ]
                    }
                    logging.info(
                        f'Adding "{name}" to applications for this AccessRules.'
                    )
            else:
                logging.warning(
                    f'Application: "{name}", ' f"not found.  Cannot add to AccessRules."
                )
        elif action == "addappfilter":
            app = ApplicationFilters(fmc=self.fmc)
            app.get(name=name)
            if "id" in app.__dict__:
                if "applicationFilters" in self.__dict__:
                    new_app = {"name": app.name, "id": app.id, "type": app.type}
                    duplicate = False
                    if "applicationFilters" not in self.applications:
                        self.__dict__["applicationFilters"]["applicationFilters"] = []
                    for obj in self.applications["applicationFilters"]:
                        if obj["name"] == new_app["name"]:
                            duplicate = True
                            break
                    if not duplicate:
                        self.applications["applicationFilters"].append(new_app)
                        logging.info(
                            f'Adding "{name}" to applications for this AccessRules.'
                        )
                else:
                    self.applications = {
                        "applicationFilters": [
                            {"name": app.name, "id": app.id, "type": app.type}
                        ]
                    }
                    logging.info(
                        f'Adding "{name}" application filter to applications for this AccessRules.'
                    )
            else:
                logging.warning(
                    f'Application Filter: "{name}", '
                    f"not found.  Cannot add to AccessRules."
                )
        elif action == "remove":
            app = Applications(fmc=self.fmc)
            app.get(name=name)
            if "id" in app.__dict__:
                if "applicationFilters" in self.__dict__:
                    applications = []
                    for obj in self.applications["applications"]:
                        if obj["name"] != name:
                            applications.append(obj)
                    self.applications["applicationFilters"] = applications
                    logging.info(
                        f'Removed "{name}" from applications for this AccessRules.'
                    )
                else:
                    logging.info(
                        "Application doesn't exist for this AccessRules.  Nothing to remove."
                    )
            else:
                logging.warning(
                    f"Application, {name}, not found.  Cannot remove from AccessRules."
                )
        elif action == "removeappfilter":
            app = ApplicationFilters(fmc=self.fmc)
            app.get(name=name)
            if "id" in app.__dict__:
                if "applications" in self.__dict__:
                    applications = []
                    for obj in self.applications["applicationFilters"]:
                        if obj["name"] != name:
                            applications.append(obj)
                    self.applications["applicationFilters"] = applications
                    logging.info(
                        f'Removed "{name}" application filter from applications for this AccessRules.'
                    )
                else:
                    logging.info(
                        "Application filter doesn't exist for this AccessRules.  Nothing to remove."
                    )
            else:
                logging.warning(
                    f"Application filter, {name}, not found.  Cannot remove from AccessRules."
                )

        elif action == "clear":
            if "applications" in self.__dict__:
                del self.applications
                logging.info("All Applications removed from this AccessRules object.")

    def urls_info(self, action, name=""):
        """
        Add/modify name to URLs field of AccessRules object.
        :param action: (str) 'add', 'remove', or 'clear'
        :param name: (str) Name of URLs in FMC.
        :return: None
        """
        logging.debug("In urls() for AccessRules class.")
        if action == "add":
            urlobj_json = URLs(fmc=self.fmc)
            urlobj_json.get(name=name)
            if "id" in urlobj_json.__dict__:
                item = urlobj_json
            else:
                item = URLGroups(fmc=self.fmc)
                item.get(name=name)
            if "id" in item.__dict__:
                if "urls" in self.__dict__:
                    new_url = {"name": item.name, "id": item.id}
                    duplicate = False
                    if "objects" not in self.urls:
                        self.__dict__["urls"]["objects"] = []
                    for obj in self.urls["objects"]:
                        if obj["name"] == new_url["name"]:
                            duplicate = True
                            break
                    if not duplicate:
                        self.urls["objects"].append(new_url)
                        logging.info(
                            f'Adding URLs "{name}" to URLs for this AccessRules.'
                        )
                else:
                    self.urls = {"objects": [{"name": item.name, "id": item.id}]}
                    logging.info(f'Adding URLs "{name}" to URLs for this AccessRules.')
            else:
                logging.warning(
                    f'URL Object or URL Object Group: "{name}", '
                    f"not found.  Cannot add to AccessRules."
                )
        elif action == "remove":
            urlobj_json = URLs(fmc=self.fmc)
            urlobj_json.get(name=name)
            if "id" in urlobj_json.__dict__:
                item = urlobj_json
            else:
                item = URLGroups(fmc=self.fmc)
                item.get(name=name)
            if "id" in item.__dict__:
                if "urls" in self.__dict__:
                    objects = []
                for obj in self.urls["objects"]:
                    if obj["name"] != name:
                        objects.append(obj)
                    self.urls["objects"] = objects
                    logging.info(
                        f'Removed URLs "{name}" from URLs for this AccessRules.'
                    )
                else:
                    logging.info(
                        "URLs doesn't exist for this AccessRules.  Nothing to remove."
                    )
            else:
                logging.warning(
                    f'URL Object or URL Object Group: "{name}", '
                    f"not found.  Cannot add to AccessRules."
                )
        elif action == "clear":
            if "urls" in self.__dict__:
                del self.urls
                logging.info("All URLs removed from this AccessRules object.")

    def new_comments(self, action, value):
        """
        Add a comment to the comment list
        Args:
            action (str): Add, remove or clear
            value (str): Comment value to add
        """
        if action == "add":
            self._newComments.append(value)
        if action == "remove":
            self._newComments.remove(value)
        if action == "clear":
            self._newComments = []


class Bulk(object):
    """
    Send many JSON objects in one API call.

    This is specific to the AccessRules() method.
    """

    MAX_SIZE_QTY = 1000
    MAX_SIZE_IN_BYTES = 1024000
    REQUIRED_FOR_POST = []

    def __init__(self, fmc, url=None, **kwargs):
        """
        Initialize Bulk object.

        :param fmc (object):  FMC object
        :param url (str): Base URL used for API action. If none, the URL will be
                          extracted from the first rule in the list
        :param **kwargs: Pass any/all variables for self.
        :return: None
        """
        logging.debug("In __init__() for Bulk class.")
        self.fmc = fmc
        self.items = []
        self.URL = url
        self.category = None
        self.insertBefore = None
        self.insertAfter = None
        self.section = None
        self.parse_kwargs(**kwargs)

    @property
    def URL_SUFFIX(self):
        """
        Add the URL suffixes for section, categories, insertBefore and insertAfter.

        :return (str): url
        """
        url = "?"

        if self.category:
            url = f"{url}category={self.category}&"

        if self.insertBefore:
            url = f"{url}insertBefore={self.insertBefore}&"

        if self.insertAfter:
            url = f"{url}insertAfter={self.insertAfter}&"

        if self.insertBefore and self.insertAfter:
            logging.warning("ACP rule has both insertBefore and insertAfter params")

        if self.section:
            url = f"{url}section={self.section}&"

        url = f"{url}bulk=true&"

        return url[:-1]

    def parse_kwargs(self, **kwargs):
        """
        Add/modify variables in self.

        :return: None
        """
        logging.debug("In parse_kwargs() for Bulk class.")
        if "category" in kwargs:
            self.category = kwargs["category"]

        if "insertBefore" in kwargs:
            self.insertBefore = kwargs["insertBefore"]

        if "insertAfter" in kwargs:
            self.insertAfter = kwargs["insertAfter"]

        if "section" in kwargs:
            self.section = kwargs["section"]

    def add(self, item):
        """
        :param item: (AccessRules) Add AccessRule to list of items to send to FMC.

        :return: None
        """
        self.items.append(item)
        logging.info(f"Adding {item} to bulk items list.")

    def clear(self):
        """
        Clear self.items -- Empty out list of JSON strings to send to FMC.

        :return: None
        """
        logging.info(f"Clearing bulk items list.")
        self.items = []

    def build_url(self):
        """
        Build the URL using either the URL is one was passed in or the URL
        from the first item in the list.
        """
        if self.URL:
            return self.URL

        if self.items[0].URL_SUFFIX:
            return f"{self.items[0].URL}&bulk=true"

        return f"{self.items[0].URL}{self.URL_SUFFIX}"

    def post(self):
        """
        Send list of self.items to FMC as a bulk import.

        :return: (str) requests response from FMC
        """
        url = self.build_url()

        # Provide some backward compatibility support here as with the current
        # versioning approach (non-semantic) there is no easy way signal to
        # consumers of a breaking change
        items = [
            item.format_data() if isinstance(item, AccessRules) else item
            for item in self.items
        ]

        # The FMC has two limitations for bulk operations that can be reached
        # independently. This first is the maximum amount of records. The second
        # is the maximum size in bytes of a single payload. We need to ensure
        # that if we hit either, that we post the data before continuing to the
        # next bulk operation

        data = []
        response = None

        for item in items:
            if (
                len(str(item)) + len(str(data)) >= self.MAX_SIZE_IN_BYTES
                or len(data) == self.MAX_SIZE_QTY
            ):
                logging.info(f"Posting bulk items.")
                if not self.fmc.send_to_api(method="post", url=url, json_data=data):
                    return

                data.clear()

            data.append(item)

        if data:
            logging.info(f"Posting bulk items.")
            response = self.fmc.send_to_api(method="post", url=url, json_data=data)

        return response

    def delete(self):
        """
        Delete list of self.items in FMC as a bulk operation.

        :return: (str) requests response from FMC
        """
        base_url = f"{self.build_url()}&filter=ids:"
        url = base_url

        for item in self.items:
            if len(url) + len(item.id) >= 2048:
                logging.info(f"Deleting bulk items.")
                if not self.fmc.send_to_api(method="delete", url=url[:-1]):
                    return

                url = base_url

            url = f"{url}{item.id},"

        logging.info(f"Deleting bulk items.")
        return self.fmc.send_to_api(method="delete", url=url[:-1])



import logging
import fmcapi
import time


def test__ip_range(fmc):
    logging.info("Test IPRange.  Post, get, put, delete Range Objects.")

    starttime = str(int(time.time()))
    namer = f"_fmcapi_test_{starttime}"

    obj1 = fmcapi.Ranges(fmc=fmc)
    obj1.name = namer
    obj1.value = "1.1.1.1-2.2.2.2"
    obj1.post()
    time.sleep(1)
    del obj1

    obj1 = fmcapi.Ranges(fmc=fmc, name=namer)
    obj1.get()
    obj1.value = "3.3.3.3-4.4.4.4"
    obj1.put()
    time.sleep(1)
    obj1.delete()

    logging.info("Test IPRange done.\n")



import logging
import fmcapi


def test__geolocations(fmc):
    logging.info("Testing Geolocation class. Requires a configured Geolocation")

    obj1 = fmcapi.Geolocation(fmc=fmc)
    logging.info("All Geolocation -- >")
    result = obj1.get()
    logging.info(result)
    logging.info(f"Total items: {len(result['items'])}")
    del obj1

    obj1 = fmcapi.Geolocation(fmc=fmc, name="_tmp")
    logging.info("One Geolocation -- >")
    logging.info(obj1.get())

    logging.info("Testing Geolocation class done.\n")



"""Object Services Classes."""

import logging
from .anyprotocolportobjects import AnyProtocolPortObjects
from .applications import Applications
from .applicationcategories import ApplicationCategories
from .applicationfilters import ApplicationFilters
from .applicationproductivities import ApplicationProductivities
from .applicationrisks import ApplicationRisks
from .applicationtags import ApplicationTags
from .applicationtypes import ApplicationTypes
from .certenrollments import CertEnrollments
from .continents import Continents
from .countries import Countries
from .dnsservergroups import DNSServerGroups
from .endpointdevicetypes import EndPointDeviceTypes
from .extendedaccesslist import ExtendedAccessList, ExtendedAccessListAce
from .fqdns import FQDNS
from .geolocation import Geolocation
from .icmpv4objects import ICMPv4Objects
from .icmpv6objects import ICMPv6Objects
from .ikev1ipsecproposals import IKEv1IpsecProposals
from .ikev1policies import IKEv1Policies
from .ikev2ipsecproposals import IKEv2IpsecProposals
from .ikev2policies import IKEv2Policies
from .interfacegroups import InterfaceGroups
from .interfaceobjects import InterfaceObjects
from .networkaddresses import NetworkAddresses
from .hosts import Hosts
from .networks import Networks
from .ranges import Ranges
from .isesecuritygrouptags import ISESecurityGroupTags
from .networkgroups import NetworkGroups
from .portobjectgroups import PortObjectGroups
from .ports import Ports
from .protocolportobjects import ProtocolPortObjects
from .realms import Realms
from .realmusergroups import RealmUserGroups
from .realmusers import RealmUsers
from .securitygrouptags import SecurityGroupTags
from .securityzones import SecurityZones
from .siurlfeeds import SIUrlFeeds
from .siurllists import SIUrlLists
from .slamonitors import SLAMonitors
from .tunneltags import TunnelTags
from .urls import URLs
from .urlcategories import URLCategories
from .urlgroups import URLGroups
from .usage import Usage
from .variablesets import VariableSets
from .vlangrouptags import VlanGroupTags
from .vlantags import VlanTags
from .dynamicobjects import DynamicObject
from .dynamicobjectmappings import DynamicObjectMappings
from .grouppolicies import GroupPolicies
from .timeranges import TimeRanges

logging.debug("In the object_services __init__.py file.")

__all__ = [
    "AnyProtocolPortObjects",
    "ApplicationCategories",
    "Applications",
    "ApplicationFilters",
    "ApplicationProductivities",
    "ApplicationRisks",
    "ApplicationTags",
    "ApplicationTypes",
    "CertEnrollments",
    "Continents",
    "Countries",
    "DNSServerGroups",
    "EndPointDeviceTypes",
    "ExtendedAccessList",
    "ExtendedAccessListAce",
    "FQDNS",
    "Geolocation",
    "GroupPolicies",
    "Hosts",
    "ICMPv4Objects",
    "ICMPv6Objects",
    "IKEv1IpsecProposals",
    "IKEv1Policies",
    "IKEv2IpsecProposals",
    "IKEv2Policies",
    "InterfaceGroups",
    "InterfaceObjects",
    "ISESecurityGroupTags",
    "NetworkAddresses",
    "NetworkGroups",
    "Networks",
    "PortObjectGroups",
    "Ports",
    "ProtocolPortObjects",
    "Ranges",
    "Realms",
    "RealmUserGroups",
    "RealmUsers",
    "SecurityGroupTags",
    "SecurityZones",
    "SIUrlFeeds",
    "SIUrlLists",
    "SLAMonitors",
    "TunnelTags",
    "URLCategories",
    "URLGroups",
    "URLs",
    "Usage",
    "VariableSets",
    "VlanGroupTags",
    "VlanTags",
    "DynamicObjectMappings",
    "DynamicObject",
    "TimeRanges",
]



import logging
import fmcapi
import time


def test__security_zone(fmc):
    logging.info("Test SecurityZone.  Post, get, put, delete Security Zone Objects.")

    starttime = str(int(time.time()))
    namer = f"_fmcapi_test_{starttime}"

    obj1 = fmcapi.SecurityZones(fmc=fmc)
    obj1.name = namer
    obj1.interfaceMode = "ROUTED"
    obj1.post()
    time.sleep(1)
    del obj1
    obj1 = fmcapi.SecurityZones(fmc=fmc, name=namer)
    obj1.get()
    obj1.name = "DEMO"
    obj1.put()
    time.sleep(1)
    obj1.delete()
    logging.info("Test SecurityZone done.\n")



import logging
import fmcapi
import time


def test__autonat(fmc):
    logging.info("Testing AutoNatPolicy class.")

    starttime = str(int(time.time()))
    namer = f"_fmcapi_test_{starttime}"

    # Create a container policy for FTD NAT rules
    natpol1 = fmcapi.FTDNatPolicies(fmc=fmc, name=namer)
    natpol1.post()
    natpol1.get()

    # Create original and translate objects
    obj1 = fmcapi.Networks(fmc=fmc)
    obj1.name = "_net_original"
    obj1.value = "10.0.0.0/8"
    obj1.post()
    time.sleep(1)

    obj2 = fmcapi.Hosts(fmc=fmc)
    obj2.name = "_net_xlate"
    obj2.value = "192.0.2.1"
    obj2.post()
    time.sleep(1)

    # Create identity nat object
    obj3 = fmcapi.Networks(fmc=fmc)
    obj3.name = "_net_identity"
    obj3.value = "192.168.0.0/24"
    obj3.post()
    time.sleep(1)

    # Create nat pool objects
    obj4 = fmcapi.Networks(fmc=fmc)
    obj4.name = "_net_original_pool"
    obj4.value = "172.16.0.0/24"
    obj4.post()
    time.sleep(1)

    # PAT Pool must be a range, not a subnet
    obj5 = fmcapi.Ranges(fmc=fmc)
    obj5.name = "_net_xlate_pool"
    obj5.value = "192.0.2.128-192.0.2.254"
    obj5.post()
    time.sleep(1)

    # Create interface PAT object
    obj6 = fmcapi.Networks(fmc=fmc)
    obj6.name = "_net_original_intf"
    obj6.value = "192.168.1.0/24"
    obj6.post()
    time.sleep(1)

    # Create Security Zones
    sz1 = fmcapi.SecurityZones(fmc=fmc, name="IG-INSIDE")
    sz1.post()
    sz2 = fmcapi.SecurityZones(fmc=fmc, name="SZ-OUTSIDE1")
    sz2.post()

    # Autonat a network object to a host
    autonat1 = fmcapi.AutoNatRules(fmc=fmc)
    autonat1.original_network(name="_net_original")
    autonat1.translated_network(name="_net_xlate")
    autonat1.natType = "STATIC"
    # Source and destination interface can be either an interface group or security zone
    autonat1.source_intf(name="IG-INSIDE")
    autonat1.destination_intf(name="SZ-OUTSIDE1")
    autonat1.nat_policy(name=namer)

    # Autonat identity nat
    autonat2 = fmcapi.AutoNatRules(fmc=fmc)
    autonat2.identity_nat(name="_net_identity")
    autonat2.source_intf(name="IG-INSIDE")
    autonat2.destination_intf(name="SZ-OUTSIDE1")
    autonat2.nat_policy(name=namer)

    # Autonat nat pool
    autonat3 = fmcapi.AutoNatRules(fmc=fmc)
    autonat3.original_network(name="_net_original_pool")
    autonat3.patPool(name="_net_xlate_pool")
    autonat3.source_intf(name="IG-INSIDE")
    autonat3.destination_intf(name="SZ-OUTSIDE1")
    autonat3.nat_policy(name=namer)

    # Autonat interface PAT
    autonat4 = fmcapi.AutoNatRules(fmc=fmc)
    autonat4.original_network(name="_net_original_intf")
    autonat4.natType = "DYNAMIC"
    autonat4.source_intf(name="IG-INSIDE")
    autonat4.destination_intf(name="SZ-OUTSIDE1")
    autonat4.nat_policy(name=namer)
    autonat4.interfaceInTranslatedNetwork = True

    autonat1.post()
    autonat2.post()
    autonat3.post()
    autonat4.post()

    """
    # Associate a nat policy to a device
    # Do not uncomment if you do not have a device registered to FMC
    # Use name of device or deviceHAPair as applicable
    pol_devices = [{"name": "deviceHAName", "type": "deviceHAPair"}]
    assign1 = fmcapi.PolicyAssignments(fmc=fmc)
    assign1.ftd_natpolicy(name=namer, devices=pol_devices)
    assign1.post()
    assign1.ftd_natpolicy(name=namer, devices=[])
    assign1.put()
    """

    natpol1.delete()
    obj1.delete()
    obj2.delete()
    obj3.delete()
    obj4.delete()
    obj5.delete()
    obj6.delete()
    sz1.delete()
    sz2.delete()



import logging
import fmcapi


def test__continent(fmc):
    logging.info("Testing Continent class.")

    obj1 = fmcapi.Continents(fmc=fmc)
    logging.info("All Continents -- >")
    result = obj1.get()
    logging.info(result)
    logging.info(f"Total items: {len(result['items'])}")
    logging.info("\n")
    del obj1
    obj1 = fmcapi.Continents(fmc=fmc, name="North America")
    logging.info("One Continent -- >")
    logging.info(obj1.get())
    logging.info("\n")

    logging.info("Testing Continent class done.\n")



import logging
import fmcapi


def test__url_category(fmc):
    logging.info("Testing URLCategory class.")
    obj1 = fmcapi.URLCategories(fmc=fmc)
    logging.info("All URLCategories -- >")
    result = obj1.get()
    logging.info(result)
    logging.info(f"Total items: {len(result['items'])}")

    del obj1
    obj1 = fmcapi.URLCategories(fmc=fmc, name="SPAM URLs")
    logging.info("One URLCategory -- >")
    logging.info(obj1.get())

    logging.info("Testing URLCategory class done.\n")



"""MonitoredInterfaces Classes."""

from fmcapi.api_objects.apiclasstemplate import APIClassTemplate
from .ftddevicehapairs import FTDDeviceHAPairs
import logging


class MonitoredInterfaces(APIClassTemplate):
    """The MonitoredInterfaces Object in the FMC."""

    VALID_JSON_DATA = [
        "id",
        "name",
        "ipv4Configuration",
        "ipv6Configuration",
        "monitorForFailures",
    ]
    VALID_FOR_KWARGS = VALID_JSON_DATA + ["ha_name"]
    PREFIX_URL = "/devicehapairs/ftddevicehapairs"
    REQUIRED_FOR_PUT = ["id"]

    def __init__(self, fmc, **kwargs):
        """
        Initialize MonitoredInterfaces object.

        :param fmc (object): FMC object
        :param **kwargs: Any other values passed during instantiation.
        :return: None
        """
        super().__init__(fmc, **kwargs)
        logging.debug("In __init__() for MonitoredInterfaces class.")
        self.parse_kwargs(**kwargs)

    def parse_kwargs(self, **kwargs):
        """
        Parse the kwargs and set self variables to match.

        :return: None
        """
        super().parse_kwargs(**kwargs)
        logging.debug("In parse_kwargs() for MonitoredInterfaces class.")
        if "ha_name" in kwargs:
            self.device_ha(ha_name=kwargs["ha_name"])

    def device_ha(self, ha_name):
        """
        Add deviceha_id to URL.

        :param ha_name: (str) Name of Device HA.
        :return: None
        """
        logging.debug("In device_ha() for MonitoredInterfaces class.")
        deviceha1 = FTDDeviceHAPairs(fmc=self.fmc, name=ha_name)
        deviceha1.get()
        if "id" in deviceha1.__dict__:
            self.deviceha_id = deviceha1.id
            self.URL = f"{self.fmc.configuration_url}{self.PREFIX_URL}/{self.deviceha_id}/monitoredinterfaces"
            self.deviceha_added_to_url = True
        else:
            logging.warning(
                f"Device HA {ha_name} not found.  Cannot set up device for MonitoredInterfaces."
            )

    def ipv4(self, ipv4addr, ipv4mask, ipv4standbyaddr):
        """
        IPv4 address info for monitored device.

        :param ipv4addr: (str) x.x.x.x
        :param ipv4mask: (str) bitmask
        :param ipv4standbyaddr: (str) x.x.x.x
        """
        logging.debug("In ipv4() for MonitoredInterfaces class.")
        self.ipv4Configuration = {
            "activeIPv4Address": ipv4addr,
            "activeIPv4Mask": ipv4mask,
            "standbyIPv4Address": ipv4standbyaddr,
        }

    def post(self):
        """POST method for API for MonitoredInterfaces not supported."""
        logging.info("POST method for API for MonitoredInterfaces not supported.")
        pass



import logging
import fmcapi
import time


def test__ipv6staticroutes(fmc):
    logging.info(
        "Testing IPv6StaticRoutes class. get, post, put, delete IPv4StaticRoute Objects. Requires a registered device"
    )

    starttime = str(int(time.time()))
    namer = f"_fmcapi_test_{starttime}"

    iphost1 = fmcapi.Hosts(fmc=fmc, name="_iphost1" + namer, value="2001:db8::1")
    iphost1.post()
    ipnet1 = fmcapi.Networks(fmc=fmc, name="_ipnet1" + namer, value="2001:db8:1::/64")
    ipnet2 = fmcapi.Networks(fmc=fmc, name="_ipnet2" + namer, value="2001:db8:2::/64")
    ipnet1.post()
    ipnet2.post()

    ipv4route1 = fmcapi.IPv6StaticRoutes(fmc=fmc, name="_ipv6route1")
    ipv4route1.device(device_name="ftdv01.ccie.lab")
    ipv4route1.networks(action="add", networks=[ipnet1.name, ipnet2.name])
    ipv4route1.gw(name=iphost1.name)
    ipv4route1.interfaceName = "ifname"
    ipv4route1.metricValue = 1
    result = ipv4route1.post()

    ipv4route2 = fmcapi.IPv4StaticRoutes(fmc=fmc, name="_ipv6route1")
    ipv4route2.device(device_name="device_name")
    ipv4route2.id = result["id"]
    ipv4route2.get()

    del ipv4route1
    ipv4route2.networks(action="remove", networks=[ipnet2.name])
    ipv4route2.put()

    ipv4route2.delete()
    ipnet1.delete()
    ipnet2.delete()
    iphost1.delete()
    logging.info("Testing IPv6StaticRoutes class done.\n")



import logging
import fmcapi


def test__dynamic_objects(fmc):
    logging.info("Testing Dynamic Objects.")
    result = fmcapi.DynamicObject(fmc=fmc).get()
    logging.info(f"All Dynamic Objects ---> {result}")
    logging.info(f"Total items: {len(result['items'])}")

    obj1 = fmcapi.DynamicObject(fmc=fmc, name="Platform",
    description="IPs of Engineer department",
    type="DynamicObject", objectType="IP",).post()

    obj1 = fmcapi.DynamicObject(fmc=fmc, name="NWS")
    obj1.get()
    logging.info(f"One objectby name----> {obj1}")




import logging
import fmcapi
import time
import json


def test__usage(fmc):
    logging.info("Test Object Find Usage.")

    starttime = str(int(time.time()))
    namer = f"_fmcapi_test_{starttime}"

    obj1 = fmcapi.Hosts(fmc=fmc)
    obj1.name = namer
    obj1.value = "8.8.8.8/32"
    obj1.post()
    time.sleep(1)
    obj2 = fmcapi.Usage(fmc=fmc)
    obj2.get(uuid=obj1.id, type=obj1.type)
    time.sleep(1)
    obj1.delete()
    del obj1
    del obj2

    logging.info("Test Object Find Usage done.\n")


"""DynamicAccessPolicies Class."""

from fmcapi.api_objects.apiclasstemplate import APIClassTemplate
import logging
import xmltodict
import base64


class DynamicAccessPolicies(APIClassTemplate):
    """The DynamicAccessPolicies Object in the FMC."""

    REQUIRED_FOR_PUT = [
        "id",
        "authorizationAttributes",
        "dapXmlConfig_dict",
        "hostscanXmlConfig_dict",
    ]
    VALID_JSON_DATA = [
        "id",
        "name",
        "type",
        "authorizationAttributes",
        "dapXmlConfig",
        "hostscanXmlConfig",
    ]
    VALID_FOR_KWARGS = VALID_JSON_DATA + []
    URL_SUFFIX = "/policy/dynamicaccesspolicies"
    FIRST_SUPPORTED_FMC_VERSION = "7.2"

    def __init__(self, fmc, **kwargs):
        """
        Initialize DynamicAccessPolicies object.

        :param fmc: (object) FMC object
        :param kwargs: Any other values passed during instantiation.
        :return: None
        """
        super().__init__(fmc, **kwargs)
        logging.debug("In __init__() for DynamicAccessPolicies class.")
        self.type = "DynamicAccessPolicy"
        self.parse_kwargs(**kwargs)

    def unpack_xml_config(self, encoded_xml):
        """
        Unpack and reformat base64 encoded xml from DAP api responses

        Args:
            encoded_xml (string): base64 encoded xml string from either dapXmlConfig or hostscanXmlConfig

        Returns:
            dict: base64 decoded 'encoded_xml' represented as a dict instead of xml
        """

        decoded_xml = base64.b64decode(encoded_xml).decode("utf-8")
        xml_dict = xmltodict.parse(decoded_xml)

        return xml_dict

    def repack_xml_config(self, xml_dict):
        """Repack dict into xml and encoded for use with DAP api calls

        Args:
            xml_dict (_type_): _description_

        Returns:
            _type_: _description_
        """

        xml = xmltodict.unparse(xml_dict)
        encoded_xml_bytes = base64.b64encode(xml.encode("utf-8"))
        encoded_xml_string = encoded_xml_bytes.decode("utf-8")

        return encoded_xml_string

    def get(self, **kwargs):
        """
        Modified get() for DynamicAccessPolicies class to make working with
        dapXmlConfig and hostscanXmlConfig a bit easier. This will create self.dapXmlConfig_dict
        and self.hostscanXmlConfig_dict from the base64 encoded xml from the api repsonse.
        """

        super().get(**kwargs)
        logging.debug("In get() for DynamicAccessPolicies class.")
        if hasattr(self, "dapXmlConfig"):
            self.dapXmlConfig_dict = self.unpack_xml_config(self.dapXmlConfig)

        if hasattr(self, "hostscanXmlConfig"):
            self.hostscanXmlConfig_dict = self.unpack_xml_config(self.hostscanXmlConfig)

    def post(self, **kwargs):
        """
        Modified post() for DynamicAccessPolicies class to make working with
        dapXmlConfig and hostscanXmlConfig a bit easier. This will take self.dapXmlConfig_dict
        and self.hostscanXmlConfig_dict and convert to xml before base64 encoding for use with api endpoint.
        """

        if hasattr(self, "dapXmlConfig_dict"):
            self.dapXmlConfig = self.repack_xml_config(self.dapXmlConfig_dict)

        if hasattr(self, "hostscanXmlConfig_dict"):
            self.hostscanXmlConfig = self.repack_xml_config(self.hostscanXmlConfig_dict)

        super().post(**kwargs)
        logging.debug("In post() for DynamicAccessPolicies class.")

    def put(self, **kwargs):
        """
        Modified put() for DynamicAccessPolicies class to make working with
        dapXmlConfig and hostscanXmlConfig a bit easier. This will take self.dapXmlConfig_dict
        and self.hostscanXmlConfig_dict and convert to xml before base64 encoding for use with api endpoint.
        """

        if hasattr(self, "dapXmlConfig_dict"):
            self.dapXmlConfig = self.repack_xml_config(self.dapXmlConfig_dict)

        if hasattr(self, "hostscanXmlConfig_dict"):
            self.hostscanXmlConfig = self.repack_xml_config(self.hostscanXmlConfig_dict)

        super().put(**kwargs)
        logging.debug("In put() for DynamicAccessPolicies class.")



"""PreFilter Policies Class."""

from fmcapi.api_objects.apiclasstemplate import APIClassTemplate
import logging


class PreFilterPolicies(APIClassTemplate):
    """The PreFilterPolicies Object in the FMC."""

    VALID_JSON_DATA = ["id", "name", "type", "description", "defaultAction"]
    VALID_FOR_KWARGS = VALID_JSON_DATA + []
    URL_SUFFIX = "/policy/prefilterpolicies"
    VALID_CHARACTERS_FOR_NAME = """[.\w\d_\- ]"""
    REQUIRED_FOR_POST = ["name"]
    DEFAULT_ACTION_OPTIONS = ["ANALYZE_TUNNELS", "BOCK_TUNNELS"]
    FIRST_SUPPORTED_FMC_VERSION = "6.5"

    def __init__(self, fmc, **kwargs):
        """
        Initialize PreFilterPolicies object.

        Set self.type to "PreFilterPolicy" and parse the kwargs.

        :param fmc (object): FMC object
        :param **kwargs: Any other values passed during instantiation.
        :return: None
        """
        super().__init__(fmc, **kwargs)
        logging.debug("In __init__() for PreFilterPolicies class.")
        self.parse_kwargs(**kwargs)
        self.type = "PreFilterPolicy"
        self._defaultAction = None
        self.defaultAction = "ANALYZE_TUNNELS"

    @property
    def defaultAction(self):
        """
        Getter for defaultAction.

        :return: {"type": "PrefilterPolicyDefaultAction", "action": self._defaultAction}
        """
        return {"type": "PrefilterPolicyDefaultAction", "action": self._defaultAction}

    @defaultAction.setter
    def defaultAction(self, action):
        """Setter for defaultAction."""
        if action in self.DEFAULT_ACTION_OPTIONS:
            self._defaultAction = action
        else:
            logging.error(
                f"action, {action}, is not a valid option.  Choose from {self.DEFAULT_ACTION_OPTIONS}."
            )

    def format_data(self):
        """
        Gather all the data in preparation for sending to API in JSON format.

        :param filter_query: (str) 'all' or 'kwargs'
        :return: (dict) json_data
        """
        json_data = super().format_data()
        logging.debug("In format_data() for AccessPolicies class.")
        json_data["defaultAction"] = self.defaultAction
        return json_data

    def put(self):
        """PUT method for API for PreFilterPolicies not supported."""
        logging.info("PUT method for API for PreFilterPolicies not supported.")
        pass



"""AccessPolicy Category Class."""

from fmcapi.api_objects.apiclasstemplate import APIClassTemplate
from fmcapi.api_objects.policy_services.accesspolicies import AccessPolicies
import logging


class Categories(APIClassTemplate):
    """
    The AccessPolicy Categories Object in the FMC.
    """

    VALID_JSON_DATA = [
        "id",
        "name",
        "type",
    ]
    VALID_FOR_KWARGS = VALID_JSON_DATA + [
        "acp_id",
        "acp_name",
        "aboveCategory",
        "insertBefore",
        "insertAfter",
        "section",
    ]
    FIRST_SUPPORTED_FMC_VERSION = "6.5"
    PREFIX_URL = "/policy/accesspolicies"
    REQUIRED_FOR_POST = ["name", "acp_id"]
    REQUIRED_FOR_GET = ["acp_id"]
    VALID_CHARACTERS_FOR_NAME = """[.\w\d_\-<>, ]"""

    @property
    def URL_SUFFIX(self):
        """
        Add the URL suffixes for aboveCategory, insertBefore and insertAfter
        NOTE: You must specify these at the time the object is initialized (created) for this feature
        to work correctly. Example:
            This works:
                new_category = Categories(fmc=fmc, acp_name='acp1', insertBefore=2)

            This does not:
                new_category = Categories(fmc=fmc, acp_name='acp1')
                new_category.insertBefore = 2
        """
        url = "?"

        if "aboveCategory" in self.__dict__:
            url = f"{url}aboveCategory={self.aboveCategory}&"
        if "insertBefore" in self.__dict__:
            url = f"{url}insertBefore={self.insertBefore}&"
        if "insertAfter" in self.__dict__:
            url = f"{url}insertAfter={self.insertAfter}&"
        if "insertBefore" in self.__dict__ and "insertAfter" in self.__dict__:
            logging.warning("ACP category has both insertBefore and insertAfter params")
        if "section" in self.__dict__:
            url = f"{url}section={self.section}&"

        return url[:-1]

    def __init__(self, fmc, **kwargs):
        """
        Initialize Categories object.

        Set self.type to "Category", parse the kwargs, and set up the self.URL.

        :param fmc: (object) FMC object
        :param kwargs: Any other values passed during instantiation.
        :return: None
        """
        super().__init__(fmc, **kwargs)
        logging.debug("In __init__() for Categories class.")
        self.type = "Category"
        self.parse_kwargs(**kwargs)
        self.URL = f"{self.URL}{self.URL_SUFFIX}"

    def parse_kwargs(self, **kwargs):
        """
        Parse the kwargs and set self variables to match.

        :return: None
        """
        super().parse_kwargs(**kwargs)
        logging.debug("In parse_kwargs() for Categories class.")
        if "acp_id" in kwargs:
            self.acp(id=kwargs["acp_id"])
        if "acp_name" in kwargs:
            self.acp(name=kwargs["acp_name"])

    def acp(self, name="", id=""):
        """
        Associate an AccessPolicies object with this Categories object.

        :param name: (str)  Name of ACP.
        :param id: (str) ID of ACP.
        :return: None
        """
        # either name or id of the ACP should be given
        logging.debug("In acp() for Categories class.")
        if id != "":
            self.acp_id = id
            self.URL = f"{self.fmc.configuration_url}{self.PREFIX_URL}/{id}/categories"
        elif name != "":
            acp1 = AccessPolicies(fmc=self.fmc)
            acp1.get(name=name)
            if "id" in acp1.__dict__:
                self.acp_id = acp1.id
                self.URL = f"{self.fmc.configuration_url}{self.PREFIX_URL}/{acp1.id}/categories"
            else:
                logging.warning(
                    f"Access Control Policy {name} not found.  Cannot set up accessPolicy for Categories."
                )
        else:
            logging.error("No accessPolicy name or ID was provided.")



"""Moving the fmc.auditrecords to an actual api_object."""

from fmcapi.api_objects.apiclasstemplate import APIClassTemplate
import logging


class AuditRecords(APIClassTemplate):
    """The AuditRecords Object in the FMC."""

    VALID_JSON_DATA = []
    VALID_FOR_KWARGS = VALID_JSON_DATA + [
        "id",
        "username",
        "subsystem",
        "source",
        "starttime",
        "endtime",
        "limit",
    ]
    URL_SUFFIX = "/audit/auditrecords"

    def __init__(self, fmc, **kwargs):
        """
        Initialize AuditRecords object.

        :param fmc (object): FMC object
        :param **kwargs: Any other values passed during instantiation.
        :return: requests response
        """
        super().__init__(fmc, **kwargs)
        logging.debug("In __init__() for AuditRecords class.")
        self.url_parameters = ""
        self.URL = f"{self.fmc.platform_url}/domain/{self.fmc.uuid}{self.URL_SUFFIX}"
        self.parse_kwargs(**kwargs)

    def parse_kwargs(self, **kwargs):
        """
        Parse the kwargs and set self variables to match.

        :return: None
        """
        logging.debug("In parse_kwargs() for AuditRecords class.")
        tmp = self.url_parameters
        if "username" in kwargs:
            self.url_parameters += f"username={kwargs['username']}&"
        if "subsystem" in kwargs:
            self.url_parameters += f"subsystem={kwargs['subsystem']}&"
        if "source" in kwargs:
            self.url_parameters += f"source={kwargs['source']}&"
        if "starttime" in kwargs:
            self.url_parameters += f"starttime={kwargs['starttime']}&"
        if "endtime" in kwargs:
            self.url_parameters += f"endtime={kwargs['endtime']}&"
        if "limit" in kwargs:
            self.url_parameters += f"limit={kwargs['limit']}&"
        if tmp is not self.url_parameters:
            self.url_parameters = self.url_parameters[:-1]
        self.URL += self.url_parameters

    def get(self):
        """
        Submit GET to FMC.

        This API function supports filtering the GET query URL with: username, subsystem, source, starttime, and
        endtime parameters.

        :return: response
        """
        logging.debug("GET method for API for AuditRecords.")
        return self.fmc.send_to_api(method="get", url=self.URL)

    def post(self):
        """POST method for API for AuditRecords not supported."""
        logging.info("POST method for API for AuditRecords not supported.")
        pass

    def put(self):
        """PUT method for API for AuditRecords not supported."""
        logging.info("PUT method for API for AuditRecords not supported.")
        pass

    def delete(self):
        """DELETE method for API for AuditRecords not supported."""
        logging.info("DELETE method for API for AuditRecords not supported.")
        pass



import logging
import fmcapi


def test__country(fmc):
    logging.info("Testing Country class.")
    obj1 = fmcapi.Countries(fmc=fmc)
    logging.info("All Countries -- >")
    result = obj1.get()
    logging.info(result)
    logging.info(f"Total items: {len(result['items'])}")
    del obj1
    obj1 = fmcapi.Countries(fmc=fmc, name="Isle Of Man")
    logging.info("One Country -- >")
    logging.info(obj1.get())

    logging.info("Testing Country class done.\n")



import logging
import fmcapi
import time


def test__devicegrouprecords(fmc):
    logging.info(
        "Test DeviceGroupRecords: "
        "get, post, put, delete DeviceGroupRecords Objects requires registered device"
    )

    starttime = str(int(time.time()))
    namer = f"_fmcapi_test_{starttime}"

    device_list = [{"name": "ftdv-HA", "type": "deviceHAPair"}]
    dg1 = fmcapi.DeviceGroupRecords(fmc=fmc)
    dg1.name = "_dg1" + namer
    dg1.devices(action="add", members=device_list)
    dg1.post()
    time.sleep(1)

    dg1.get()
    dg1.devices(action="remove", members=device_list)
    dg1.put()
    time.sleep(1)

    dg1.get()
    dg1.devices(action="add", members=device_list)
    dg1.put()
    time.sleep(1)

    dg1.get()
    dg1.devices(action="clear")
    dg1.put()
    time.sleep(1)

    dg1.get()
    dg1.delete()
    logging.info("Testing DeviceGroupRecords class done.\n")



"""List Applicable Devices Class."""

from fmcapi.api_objects.apiclasstemplate import APIClassTemplate
from .upgradepackages import UpgradePackages
import logging


class ListApplicableDevices(APIClassTemplate):
    """The ListApplicableDevices Object in the FMC."""

    VALID_JSON_DATA = [
        "id",
        "name",
        "type",
        "upgradePackage",
        "model",
        "modelId",
        "modelNumber",
        "modelType",
        "healthStatus",
        "sw_version",
        "isPartofContainer",
        "containerType",
        "healthPolicy",
        "accessPolicy",
    ]
    VALID_FOR_KWARGS = VALID_JSON_DATA + []
    URL_SUFFIX = "/updates/upgradepackages"

    def __init__(self, fmc, **kwargs):
        """
        Initialize ListApplicableDevices object.

        Set self.type to "UpgradePackage", and parse the kwargs.
        :param fmc (object): FMC object
        :param **kwargs: Any other values passed during instantiation.
        :return: None
        """
        super().__init__(fmc, **kwargs)
        logging.debug("In __init__() for ListApplicableDevices class.")
        self.type = "UpgradePackage"
        self.parse_kwargs(**kwargs)

    def upgrade_package(self, package_name):
        """
        Upgrade named package.

        :param package_name: (str) Name of package to upgrade
        :return: None
        """
        logging.debug("In upgrade_package() for ListApplicableDevices class.")
        package1 = UpgradePackages(fmc=self.fmc)
        package1.get(name=package_name)
        if "id" in package1.__dict__:
            self.package_id = package1.id
            self.URL = f"{self.fmc.platform_url}{self.URL_SUFFIX}/{self.package_id}/applicabledevices"
            self.package_added_to_url = True
        else:
            logging.warning(
                f"UpgradePackage {package_name} not found.  Cannot get list of ListApplicableDevices."
            )

    def post(self):
        """POST method for API for ListApplicableDevices not supported."""
        logging.info("POST method for API for ListApplicableDevices not supported.")
        pass

    def put(self):
        """PUT method for API for ListApplicableDevices not supported."""
        logging.info("PUT method for API for ListApplicableDevices not supported.")
        pass

    def delete(self):
        """DELETE method for API for ListApplicableDevices not supported."""
        logging.info("DELETE method for API for ListApplicableDevices not supported.")
        pass



"""Device Records Classes."""

from fmcapi.api_objects.apiclasstemplate import APIClassTemplate
from fmcapi.api_objects.policy_services.accesspolicies import AccessPolicies
from fmcapi.api_objects.status_services import TaskStatuses
import time
import logging


class DeviceRecords(APIClassTemplate):
    """The DeviceRecords Object in the FMC."""

    VALID_JSON_DATA = [
        "id",
        "name",
        "type",
        "hostName",
        "natID",
        "regKey",
        "license_caps",
        "performanceTier",
        "accessPolicy",
    ]
    VALID_GET_FILTERS = [
        # TODO: name as a get filter will conflict with existing logic to look something up by name - use hostname instead for now
        # "name", # String
        "hostName",  # String
        "serialNumber",  # String
        "containerType",  # DeviceCluster|DeviceHAPair|DeviceStack
        "clusterBootstrapSupported",  # Bool
        "analyticsOnly",  # Bool
        "includeOtherAssociatedPolicies",  # Bool
        "modeType",  # NGFW|Chassis
    ]
    VALID_FOR_KWARGS = (
        VALID_JSON_DATA
        + VALID_GET_FILTERS
        + [
            "acp_name",
            "acp_id",
            "model",
            "modelId",
            "modelNumber",
            "modelType",
            "healthStatus",
            "healthPolicy",
            "type",
            "version",
            "sw_version",
            "deviceGroup",
            "prohibitPacketTransfer",
            "keepLocalEvents",
            "ftdMode",
            "keepLocalEvents",
            "expanded",
        ]
    )
    URL_SUFFIX = "/devices/devicerecords"
    REQUIRED_FOR_POST = ["accessPolicy", "hostName", "regKey", "type"]
    REQUIRED_FOR_PUT = ["id"]
    LICENSES = [
        "BASE",
        "THREAT",
        "URLFilter",
        "MALWARE",
        "APEX",
        "PLUS",
        "VPNOnly",
        "INSTANCE",
    ]
    TIERS = ["FTDv5", "FTDv10", "FTDv20", "FTDv30", "FTDv50", "FTDv100", "Legacy"]

    def __init__(self, fmc, **kwargs):
        """
        Initialize DeviceRecords object.

        :param fmc (object): FMC object
        :param **kwargs: Any other values passed during instantiation.
        :return: None
        """
        super().__init__(fmc, **kwargs)
        logging.debug("In __init__() for DeviceRecords class.")
        self.parse_kwargs(**kwargs)

    def parse_kwargs(self, **kwargs):
        """
        Parse the kwargs and set self variables to match.

        :return: None
        """
        super().parse_kwargs(**kwargs)
        logging.debug("In parse_kwargs() for DeviceRecords class.")
        if "acp_name" in kwargs:
            self.acp(name=kwargs["acp_name"])

    def licensing(self, action, name="BASE"):
        """
        Associate licenses with this device record.

        :param action: (str) 'add', 'remove', 'clear'
        :param name: (str) Value from LICENSES constant.
        :return: None
        """
        logging.debug("In licensing() for DeviceRecords class.")
        if action == "add":
            if name in self.LICENSES:
                if "license_caps" in self.__dict__:
                    self.license_caps.append(name)
                    self.license_caps = list(set(self.license_caps))
                else:
                    self.license_caps = [name]
                logging.info(f'License "{name}" added to this DeviceRecords object.')

            else:
                logging.warning(
                    f"{name} not found in {self.LICENSES}.  Cannot add license to DeviceRecords."
                )
        elif action == "remove":
            if name in self.LICENSES:
                if "license_caps" in self.__dict__:
                    try:
                        self.license_caps.remove(name)
                    except ValueError:
                        logging.warning(
                            f"{name} is not assigned to this devicerecord thus cannot be removed."
                        )
                    logging.info(
                        f'License "{name}" removed from this DeviceRecords object.'
                    )
                else:
                    logging.warning(
                        f"{name} is not assigned to this devicerecord thus cannot be removed."
                    )

            else:
                logging.warning(
                    f"{name} not found in {self.LICENSES}.  Cannot remove license from DeviceRecords."
                )
        elif action == "clear":
            if "license_caps" in self.__dict__:
                del self.license_caps
                logging.info("All licensing removed from this DeviceRecords object.")

    def tiering(self, action, name=""):
        """
        Associate performance tier with this device record.

        :param action: (str) 'add', 'remove', 'clear'
        :param name: (str) Value from TIERS constant.
        :return: None
        """
        logging.debug("In tiering() for DeviceRecords class.")
        if self.fmc.serverVersion < "7.0":
            logging.warning(
                f"FTD performance tier licenses are supported only in FMC version 7.0 and newer."
            )
        else:
            if action == "add":
                if name in self.TIERS:
                    self.performanceTier = name
                    logging.info(
                        f'Performance tier "{name}" added to this DeviceRecords object.'
                    )
                else:
                    logging.warning(
                        f"{name} not found in {self.TIERS}.  Cannot add performance tier to DeviceRecords."
                    )
            elif action == "remove":
                if name in self.TIERS:
                    if "performanceTier" in self.__dict__:
                        try:
                            self.performanceTier = ""
                        except ValueError:
                            logging.warning(
                                f"{name} performance tier cannot be removed."
                            )
                        logging.info(
                            f'License "{name}" removed from this DeviceRecords object.'
                        )
                    else:
                        logging.warning(
                            f"{name} is not assigned to this DeviceRecords thus cannot be removed."
                        )

                else:
                    logging.warning(
                        f"{name} not found in {self.TIERS}.  Cannot remove performance tier from DeviceRecords."
                    )
            elif action == "clear":
                if "performanceTier" in self.__dict__:
                    del self.performanceTier
                    logging.info(
                        "Performance tier removed from this DeviceRecords object."
                    )

    def acp(self, name=""):
        """
        Associate AccessPolicy with this device.

        :param name: (str) Name of ACP.
        :return: None
        """
        logging.debug("In acp() for DeviceRecords class.")
        acp = AccessPolicies(fmc=self.fmc)
        acp.get(name=name)
        if "id" in acp.__dict__:
            self.accessPolicy = {"id": acp.id, "type": acp.type}
        else:
            logging.warning(
                f"Access Control Policy {name} not found.  Cannot set up accessPolicy for DeviceRecords."
            )

    def wait_for_task(self, task, wait_time=10):
        """
        Pause configuration script and wait for device registration to complete.

        :param task: (dict) task["id": (str)]
        :param wait_time: (int) Seconds to wait before rechecking.
        :return: None
        """
        task_completed_states = ["Success", "SUCCESS", "COMPLETED"]
        try:
            status = TaskStatuses(fmc=self.fmc, id=task["id"])
            current_status = status.get()
            """
            Task Status for new device registration behaves differently than other tasks
            On new device registration, a task is sent for the initial registration. After completion
            the UUID is deleted without any change in task status. So we check to see if the object no longer exists
            to assume the registration is complete.  After registration, discovery of the device begins, but there is
            no way to check for this with a task status.  The device can't be modified during this time, but a new
            device registration can begin.

            OTOH, a device HA operation will update its status to "Success" on completion.  Hence the two different
            checks.
            """
            while (
                current_status["status"] is not None
                and current_status["status"] not in task_completed_states
            ):
                # Lot of inconsistencies with the type of data a task can return
                if "taskType" in current_status.keys():
                    logging.info(
                        f"Task: {current_status['taskType']} {current_status['status']} {current_status['id']}"
                    )
                    time.sleep(wait_time)
                    current_status = status.get()
                else:
                    logging.info(
                        f"Task: {current_status['status']} {current_status['id']}"
                    )
                    time.sleep(wait_time)
                    current_status = status.get()
            logging.info(f"Task: {current_status['status']} {current_status['id']}")
        except Exception as e:
            logging.info(type(e), e)

    def post(self, **kwargs):
        """POST to FMC API."""
        logging.debug("In post() for DeviceRecords class.")
        response = super().post(**kwargs)
        #  self.wait_for_task(task=response["metadata"]["task"], wait_time=30)  # Doesn't work yet.
        if "post_wait_time" in kwargs:
            self.post_wait_time = kwargs["post_wait_time"]
        else:
            self.post_wait_time = 300
        logging.info(
            f"DeviceRecords registration task submitted.  "
            f"Waiting {self.post_wait_time} seconds for it to complete."
        )
        time.sleep(self.post_wait_time)
        return response



import logging
import fmcapi
import time


def test__ftds2svpns(fmc):
    logging.info("Testing FTDS2SVPNs class.  Requires at least one registered device.")

    starttime = str(int(time.time()))
    namer = f"_fmcapi_test_{starttime}"

    # Create a Site2Site VPN Policy

    vpnpol1 = fmcapi.FTDS2SVPNs(fmc=fmc, name=namer)
    vpnpol1.topologyType = "POINT_TO_POINT"
    vpnpol1.ikeV1Enabled = True
    vpnpol1.ikeV2Enabled = False
    vpnpol1.post()
    vpnpol1.get()

    # Create some network objects for the encryption domains
    obj1 = fmcapi.Networks(fmc=fmc)
    obj1.name = "_net1_site1"
    obj1.value = "10.255.0.0/24"
    obj1.post()
    time.sleep(1)

    obj2 = fmcapi.Networks(fmc=fmc)
    obj2.name = "_net2_site1"
    obj2.value = "10.255.1.0/24"
    obj2.post()
    time.sleep(1)

    obj3 = fmcapi.Networks(fmc=fmc)
    obj3.name = "_net1_site2"
    obj3.value = "10.255.2.0/24"
    obj3.post()
    time.sleep(1)

    # Create Phase 1 settings
    # There is no way to search by name, so we just find the iksettings object inside the vpn policy
    ike1_json = fmcapi.IKESettings(fmc=fmc)
    ike1_json.vpn_policy(pol_name=namer)
    items = ike1_json.get()["items"][0]

    ike1 = fmcapi.IKESettings(fmc=fmc)
    ike1.vpn_policy(pol_name=namer)
    ike1.id = items["id"]
    ike1.get()
    ike1.ike_policy(pol_name="preshared_sha_aes192_dh5_10")

    ike1.put()

    # Create Phase 2 settings
    # There is no way to search by name, so we just find the ipsecsettings object inside the vpn policy
    ipsec1_json = fmcapi.IPSecSettings(fmc=fmc)
    ipsec1_json.vpn_policy(pol_name=namer)
    items = ipsec1_json.get()["items"][0]

    ipsec1 = fmcapi.IPSecSettings(fmc=fmc)
    ipsec1.vpn_policy(pol_name=namer)
    ipsec1.id = items["id"]
    ipsec1.get()
    ipsec1.ipsec_policy(pol_name="tunnel_aes256_sha")

    ipsec1.put()

    # Add vpn peers
    # FTD in HA mode should use the name of logical HA device
    endp1 = fmcapi.Endpoints(fmc=fmc)
    endp1.peerType = "PEER"
    endp1.connectionType = "BIDIRECTIONAL"
    endp1.vpn_policy(pol_name=namer)
    endp1.endpoint(action="add", device_name="_ha_name")
    endp1.vpn_interface(device_name="_ha_name", ifname="OUTSIDE1")
    endp1.encryption_domain(action="add", names=["_net1_site1", "_net2_site1"])

    endp2 = fmcapi.Endpoints(fmc=fmc)
    endp2.peerType = "PEER"
    endp2.connectionType = "BIDIRECTIONAL"
    endp2.vpn_policy(pol_name=namer)
    endp2.endpoint(action="add", device_name="_device_name")
    endp2.vpn_interface(device_name="device_name", ifname="OUTSIDE1")
    endp2.encryption_domain(action="add", names=["_net1_site2"])

    endp1.post()
    endp2.post()

    time.sleep(30)

    vpnpol1.delete()
    obj1.delete()
    obj2.delete()
    obj3.delete()



import logging
import fmcapi


def test__intrusion_policy(fmc):
    logging.info("Test IntrusionPolicy. Can only GET IntrusionPolicy objects.")
    obj1 = fmcapi.IntrusionPolicies(fmc=fmc)
    obj1.get(name="Security Over Connectivity")
    logging.info("IntrusionPolicy -->")
    logging.info(obj1.format_data())

    logging.info("Test IntrusionPolicy done.\n")



import logging
import fmcapi


def test__application_productivity(fmc):
    logging.info("Testing ApplicationProductivity class.")

    obj1 = fmcapi.ApplicationProductivities(fmc=fmc)
    logging.info("All ApplicationProductivities -- >")
    result = obj1.get()
    logging.info(result)
    logging.info(f"Total items: {len(result['items'])}")
    logging.info("\n")
    del obj1
    obj1 = fmcapi.ApplicationProductivities(fmc=fmc, name="Very Low")
    logging.info("One ApplicationProductivity -- >")
    logging.info(obj1.get())

    logging.info("Testing ApplicationProductivity class done.\n")



"""Networks Class."""

from fmcapi.api_objects.apiclasstemplate import APIClassTemplate
from fmcapi.api_objects.helper_functions import *
import logging


class Networks(APIClassTemplate):
    """The Networks Object in the FMC."""

    VALID_JSON_DATA = ["id", "name", "value", "description"]
    VALID_GET_FILTERS = [
        "unusedOnly",
        "nameOrValue",
    ]  # unusedOnly:Bool, nameOrValue:String
    VALID_FOR_KWARGS = VALID_JSON_DATA + VALID_GET_FILTERS + []
    URL_SUFFIX = "/object/networks"
    REQUIRED_FOR_POST = ["name", "value"]

    def __init__(self, fmc, **kwargs):
        """
        Initialize Networks object.

        :param fmc: (object) FMC object
        :param kwargs: Any other values passed during instantiation.
        :return: None
        """
        super().__init__(fmc, **kwargs)
        logging.debug("In __init__() for Networks class.")
        self.parse_kwargs(**kwargs)

    def parse_kwargs(self, **kwargs):
        """
        Parse the kwargs and set self variables to match.

        :return: None
        """
        super().parse_kwargs(**kwargs)
        logging.debug("In parse_kwargs() for AccessRules class.")
        if "value" in kwargs:
            value_type = get_networkaddress_type(kwargs["value"])
            if value_type == "range" or value_type == "host":
                logging.warning(
                    f"value, {kwargs['value']}, is of type {value_type}.  Limited functionality for this "
                    f"object due to it being created via the Networks function."
                )
            if validate_ip_bitmask_range(value=kwargs["value"], value_type=value_type):
                self.value = kwargs["value"]
            else:
                logging.error(
                    f"Provided value, {kwargs['value']}, has an error with the IP address(es)."
                )



from fmcapi.api_objects.apiclasstemplate import APIClassTemplate
import logging


class DynamicObject(APIClassTemplate):
    """The Dynamic Object in the FMC."""

    VALID_JSON_DATA = ["id", "name", "description", "type", "objectType"]
    VALID_FOR_KWARGS = VALID_JSON_DATA + []
    URL_SUFFIX = "/object/dynamicobjects"
    VALID_CHARACTERS_FOR_NAME = """[.\w\d_\- ]"""

    VALID_GET_FILTERS = [
        "unusedOnly",
        "ids" "nameStartsWith",
    ]  # unusedOnly:Bool, "ids:id1,id2,..." ,nameStartsWith:String

    def __init__(self, fmc, **kwargs):
        """
        Initialize Dynamic Object.

        Set self.type to "DynamicObject" and parse the kwargs.

        :param fmc: (object) FMC object
        :param kwargs: Any other values passed during instantiation.
        :return: None
        """
        super().__init__(fmc, **kwargs)
        logging.debug("In __init__() for Dynamic Object class.")
        self.parse_kwargs(**kwargs)
        self.type = "DynamicObject"



import logging
import fmcapi
import time


def test__network_group(fmc):
    logging.info("Testing NetworkGroup class.")

    starttime = str(int(time.time()))
    namer = f"_fmcapi_test_{starttime}"

    obj10 = fmcapi.Hosts(fmc=fmc, name="_iphost1", value="3.3.3.3")
    obj10.post()
    obj11 = fmcapi.Networks(fmc=fmc, name="_ipnet1", value="3.3.3.0/24")
    obj11.post()
    obj12 = fmcapi.Ranges(fmc=fmc, name="_iprange1", value="3.3.3.3-33.33.33.33")
    obj12.post()
    time.sleep(1)

    obj1 = fmcapi.NetworkGroups(fmc=fmc, name=namer)
    obj1.named_networks(action="add", name=obj10.name)
    obj1.named_networks(action="add", name=obj10.name)
    obj1.named_networks(action="remove", name=obj10.name)
    obj1.named_networks(action="clear")
    obj1.named_networks(action="add", name=obj11.name)
    obj1.named_networks(action="add", name=obj12.name)
    obj1.named_networks(action="remove", name=obj11.name)
    obj1.post()
    time.sleep(1)
    del obj1

    obj1 = fmcapi.NetworkGroups(fmc=fmc, name=namer)
    obj1.get()
    obj1.unnamed_networks(action="add", value="1.2.3.4")
    obj1.unnamed_networks(action="clear")
    obj1.unnamed_networks(action="add", value="1.2.3.4")
    obj1.unnamed_networks(action="remove", value="1.2.3.4")
    obj1.unnamed_networks(action="add", value="6.7.8.9")
    obj1.unnamed_networks(action="add", value="1.2.3.0/24")
    obj1.post()

    time.sleep(1)
    obj1.delete()
    obj10.delete()
    obj11.delete()
    obj12.delete()

    logging.info("Testing NetworkGroup class done.\n")



"""ICMPv4 Objects Class."""

from fmcapi.api_objects.apiclasstemplate import APIClassTemplate
import logging


class ICMPv4Objects(APIClassTemplate):
    """The ICMPv4Objects Object in the FMC."""

    VALID_JSON_DATA = [
        "id",
        "name",
        "description",
        "type",
        "overrideTargetId",
        "code",
        "icmpType",
        "overrides",
        "overridable",
    ]
    VALID_GET_FILTERS = [
        "unusedOnly",
        "nameOrValue",
    ]  # unusedOnly:Bool, nameOrValue:String
    VALID_FOR_KWARGS = VALID_JSON_DATA + VALID_GET_FILTERS + []
    URL_SUFFIX = "/object/icmpv4objects"
    VALID_CHARACTERS_FOR_NAME = """[.\w\d_\- ]"""

    def __init__(self, fmc, **kwargs):
        """
        Initialize ICMPv4Objects object.

        Set self.type to "ICMPv4Object" and parse the kwargs.

        :param fmc: (object) FMC object
        :param kwargs: Any other values passed during instantiation.
        :return: None
        """
        super().__init__(fmc, **kwargs)
        logging.debug("In __init__() for ICMPv4Objects class.")
        self.parse_kwargs(**kwargs)
        self.type = "ICMPV4Object"



import logging
import fmcapi
import time


def test__icmpv6(fmc):
    logging.info("Test ICMPv6Object.  Post, get, put, delete ICMPv6Object Objects.")

    starttime = str(int(time.time()))
    namer = f"_fmcapi_test_{starttime}"

    obj1 = fmcapi.ICMPv6Objects(fmc=fmc)
    obj1.name = "_icmpv6" + namer
    obj1.icmpType = "1"
    obj1.code = "0"
    obj1.post()

    obj1.get()
    obj1.code = "3"
    obj1.put()

    obj1.delete()

    logging.info("ICMPv6Object class done.\n")



"""Application Fitlers Class."""

from fmcapi.api_objects.apiclasstemplate import APIClassTemplate
import logging


class ApplicationFilters(APIClassTemplate):
    """The ApplicationFilters Object in the FMC."""

    VALID_JSON_DATA = [
        "id",
        "name",
        "type",
        "appConditions",
        "applications",
        "conditions",
    ]
    VALID_FOR_KWARGS = VALID_JSON_DATA + []
    URL_SUFFIX = "/object/applicationfilters"
    VALID_CHARACTERS_FOR_NAME = """[.\w\d_\- ]"""

    def __init__(self, fmc, **kwargs):
        """
        Initialize ApplicationFilters object.

        :param fmc (object): FMC object
        :param **kwargs: Any other values passed during instantiation.
        :return: None
        """
        super().__init__(fmc, **kwargs)
        logging.debug("In __init__() for ApplicationFilters class.")
        self.parse_kwargs(**kwargs)

    def post(self):
        """POST method for API for ApplicationFilters not supported."""
        logging.info("POST method for API for ApplicationFilters not supported.")
        pass

    def put(self):
        """PUT method for API for ApplicationFilters not supported."""
        logging.info("PUT method for API for ApplicationFilters not supported.")
        pass

    def delete(self):
        """DELETE method for API for ApplicationFilters not supported."""
        logging.info("DELETE method for API for ApplicationFilters not supported.")
        pass



"""Update Packages Classes."""

import logging

# TODO: Why is this the case?
from .policy_services.accesspolicies import AccessPolicies  # Needs loaded before Device
from .device_services.devicerecords import DeviceRecords  # Needs loaded early.

from .audit_services.audit_records import AuditRecords
from .backup_services.backup import Backup
from .deployment_services import DeployableDevices
from .deployment_services import DeploymentRequests
from .device_clusters import FTDDeviceCluster
from .device_group_services import DeviceGroupRecords
from .device_ha_pair_services import FTDDeviceHAPairs
from .device_ha_pair_services.failoverinterfacemacaddressconfigs import (
    FailoverInterfaceMACAddressConfigs,
)
from .device_ha_pair_services.monitoredinterfaces import MonitoredInterfaces
from .device_services.bridgegroupinterfaces import BridgeGroupInterfaces
from .device_services.etherchannelinterfaces import EtherchannelInterfaces
from .device_services.ipv4staticroutes import IPv4StaticRoutes
from .device_services.ipv6staticroutes import IPv6StaticRoutes
from .device_services.physicalinterfaces import PhysicalInterfaces
from .device_services.redundantinterfaces import RedundantInterfaces
from .device_services.staticroutes import StaticRoutes
from .device_services.subinterfaces import SubInterfaces
from .health.terminateravpnsessions import TerminateRAVPNSessions
from .object_services.anyprotocolportobjects import AnyProtocolPortObjects
from .object_services.applicationcategories import ApplicationCategories
from .object_services.applicationfilters import ApplicationFilters
from .object_services.applicationproductivities import ApplicationProductivities
from .object_services.applicationrisks import ApplicationRisks
from .object_services.applications import Applications
from .object_services.applicationtags import ApplicationTags
from .object_services.applicationtypes import ApplicationTypes
from .object_services.certenrollments import CertEnrollments
from .object_services.continents import Continents
from .object_services.countries import Countries
from .object_services.dnsservergroups import DNSServerGroups
from .object_services.endpointdevicetypes import EndPointDeviceTypes
from .object_services.extendedaccesslist import ExtendedAccessList
from .object_services.extendedaccesslist import ExtendedAccessListAce
from .object_services.fqdns import FQDNS
from .object_services.geolocation import Geolocation
from .object_services.hosts import Hosts
from .object_services.icmpv4objects import ICMPv4Objects
from .object_services.icmpv6objects import ICMPv6Objects
from .object_services.ikev1ipsecproposals import IKEv1IpsecProposals
from .object_services.ikev1policies import IKEv1Policies
from .object_services.ikev2ipsecproposals import IKEv2IpsecProposals
from .object_services.ikev2policies import IKEv2Policies
from .object_services.interfacegroups import InterfaceGroups
from .object_services.interfaceobjects import InterfaceObjects
from .object_services.isesecuritygrouptags import ISESecurityGroupTags
from .object_services.networkaddresses import NetworkAddresses
from .object_services.networkgroups import NetworkGroups
from .object_services.networks import Networks
from .object_services.portobjectgroups import PortObjectGroups
from .object_services.ports import Ports
from .object_services.protocolportobjects import ProtocolPortObjects
from .object_services.ranges import Ranges
from .object_services.realms import Realms
from .object_services.realmusergroups import RealmUserGroups
from .object_services.realmusers import RealmUsers
from .object_services.securitygrouptags import SecurityGroupTags
from .object_services.securityzones import SecurityZones
from .object_services.siurlfeeds import SIUrlFeeds
from .object_services.siurllists import SIUrlLists
from .object_services.slamonitors import SLAMonitors
from .object_services.tunneltags import TunnelTags
from .object_services.urlcategories import URLCategories
from .object_services.urlgroups import URLGroups
from .object_services.urls import URLs
from .object_services.usage import Usage
from .object_services.variablesets import VariableSets
from .object_services.vlangrouptags import VlanGroupTags
from .object_services.vlantags import VlanTags
from .policy_assignment_services.policyassignments import PolicyAssignments
from .policy_services.accessrules import AccessRules
from .policy_services.accessrules import Bulk
from .policy_services.advancedsettings import AdvancedSettings
from .policy_services.autonatrules import AutoNatRules
from .policy_services.defaultactions import DefaultActions
from .policy_services.endpoints import Endpoints
from .policy_services.filepolicies import FilePolicies
from .policy_services.ftdnatpolicies import FTDNatPolicies
from .policy_services.ftds2svpns import FTDS2SVPNs
from .policy_services.hitcounts import HitCounts
from .policy_services.ikesettings import IKESettings
from .policy_services.intrusionpolicies import IntrusionPolicies
from .policy_services.ipsecsettings import IPSecSettings
from .policy_services.loggingsettings import LoggingSettings
from .policy_services.manualnatrules import ManualNatRules
from .policy_services.natrules import NatRules
from .policy_services.prefilterpolicies import PreFilterPolicies
from .policy_services.prefilterrules import PreFilterRules
from .search.globalsearch import GlobalSearch
from .search.object import Object
from .search.policy import Policy
from .status_services.taskstatuses import TaskStatuses
from .system_information import ServerVersion
from .update_packages.listapplicabledevices import ListApplicableDevices
from .update_packages.upgradepackage import Upgrades
from .update_packages.upgradepackages import UpgradePackages
from .object_services.dynamicobjects import DynamicObject
from .object_services.dynamicobjectmappings import DynamicObjectMappings
from .object_services.grouppolicies import GroupPolicies
from .policy_services.ravpns import RAVpn
from .policy_services import ConnectionProfiles
from .policy_services import DynamicAccessPolicies
from .object_services import TimeRanges

logging.debug("In the api_objects __init__.py file.")

__all__ = [
    "AdvancedSettings",
    "IPSecSettings",
    "Endpoints",
    "FTDS2SVPNs",
    "IKESettings",
    "DeploymentRequests",
    "DeployableDevices",
    "ServerVersion",
    "AuditRecords",
    "DefaultActions",
    "DeviceGroupRecords",
    "FTDDeviceCluster",
    "FTDDeviceHAPairs",
    "FailoverInterfaceMACAddressConfigs",
    "MonitoredInterfaces",
    "DeviceRecords",
    "StaticRoutes",
    "IPv4StaticRoutes",
    "IPv6StaticRoutes",
    "PhysicalInterfaces",
    "BridgeGroupInterfaces",
    "RedundantInterfaces",
    "EtherchannelInterfaces",
    "SubInterfaces",
    "AnyProtocolPortObjects",
    "ApplicationCategories",
    "ApplicationFilters",
    "ApplicationProductivities",
    "ApplicationRisks",
    "Applications",
    "ApplicationTags",
    "ApplicationTypes",
    "CertEnrollments",
    "Continents",
    "Countries",
    "DNSServerGroups",
    "EndPointDeviceTypes",
    "ExtendedAccessList",
    "ExtendedAccessListAce",
    "FQDNS",
    "Geolocation",
    "Hosts",
    "ICMPv4Objects",
    "ICMPv6Objects",
    "IKEv1IpsecProposals",
    "IKEv1Policies",
    "IKEv2IpsecProposals",
    "IKEv2Policies",
    "InterfaceGroups",
    "InterfaceObjects",
    "ISESecurityGroupTags",
    "NetworkAddresses",
    "NetworkGroups",
    "Networks",
    "PortObjectGroups",
    "Ports",
    "ProtocolPortObjects",
    "Ranges",
    "Realms",
    "RealmUserGroups",
    "RealmUsers",
    "SecurityGroupTags",
    "SecurityZones",
    "SIUrlFeeds",
    "SIUrlLists",
    "SLAMonitors",
    "TunnelTags",
    "URLCategories",
    "URLGroups",
    "URLs",
    "VariableSets",
    "VlanGroupTags",
    "VlanTags",
    "PolicyAssignments",
    "GlobalSearch",
    "Object",
    "Policy",
    "AccessPolicies",
    "AccessRules",
    "Bulk",
    "FilePolicies",
    "FTDNatPolicies",
    "AutoNatRules",
    "ManualNatRules",
    "NatRules",
    "IntrusionPolicies",
    "PreFilterPolicies",
    "PreFilterRules",
    "HitCounts",
    "TaskStatuses",
    "ListApplicableDevices",
    "UpgradePackages",
    "Upgrades",
    "LoggingSettings",
    "Backup",
    "Usage",
    "DynamicObject",
    "DynamicObjectMappings",
    "TerminateRAVPNSessions",
    "GroupPolicies",
    "RAVpn",
    "ConnectionProfiles",
    "DynamicAccessPolicies",
    "TimeRanges",
]



"""Time Ranges Class."""

from fmcapi.api_objects.apiclasstemplate import APIClassTemplate
import logging


class TimeRanges(APIClassTemplate):
    """The TimeRanges Object in the FMC."""

    VALID_JSON_DATA = [
        "id",
        "name",
        "description",
        "effectiveStartDateTime",
        "effectiveEndDateTime",
        "recurrenceList",
    ]
    VALID_FOR_KWARGS = VALID_JSON_DATA + []
    REQUIRED_FOR_POST = ["name", "effectiveStartDateTime", "effectiveEndDateTime"]
    URL_SUFFIX = "/object/timeranges"

    def __init__(self, fmc, **kwargs):
        """
        Initialize TimeRanges object.

        :param fmc: (object) FMC object
        :param kwargs: Any other values passed during instantiation.
        :return: None
        """
        super().__init__(fmc, **kwargs)
        logging.debug("In __init__() for TimeRanges class.")
        self.parse_kwargs(**kwargs)



from fmcapi.api_objects.apiclasstemplate import APIClassTemplate
import logging


class AccessPolicies(APIClassTemplate):
    """
    The AccessPolicies Object in the FMC.
    """

    VALID_JSON_DATA = [
        "id",
        "name",
        "type",
        "description",
        "defaultAction",
        "prefilterPolicySetting",
    ]
    VALID_FOR_KWARGS = VALID_JSON_DATA + []
    VALID_CHARACTERS_FOR_NAME = """[ .\w\d_\-]"""
    URL_SUFFIX = "/policy/accesspolicies"
    REQUIRED_FOR_POST = ["name"]
    REQUIRED_FOR_PUT = ["id", "defaultAction", "defaultActionId"]
    DEFAULT_ACTION_OPTIONS = [
        "BLOCK",
        "TRUST",
        "PERMIT",
        "NETWORK_DISCOVERY",
        "INHERIT_FROM_PARENT",
    ]
    FILTER_BY_NAME = True

    @property
    def defaultAction(self):
        return {"action": self._defaultAction}

    @defaultAction.setter
    def defaultAction(self, action):
        if action in self.DEFAULT_ACTION_OPTIONS:
            self._defaultAction = action
        else:
            logging.error(
                f"action, {action}, is not a valid option.  Choose from {self.DEFAULT_ACTION_OPTIONS}."
            )

    def __init__(self, fmc, **kwargs):
        super().__init__(fmc, **kwargs)
        logging.debug("In __init__() for AccessPolicies class.")
        self._defaultAction = None
        self.prefilter_id = None
        self.prefilter_name = None
        self.defaultAction = "BLOCK"
        self.defaultActionId = None
        self.parse_kwargs(**kwargs)

    def format_data(self):
        json_data = super().format_data()
        logging.debug("In format_data() for AccessPolicies class.")

        json_data["defaultAction"] = self.defaultAction

        if self.defaultActionId:
            json_data["defaultAction"]["id"] = self.defaultActionId

        return json_data

    def parse_kwargs(self, **kwargs):
        super().parse_kwargs(**kwargs)
        logging.debug("In parse_kwargs() for AccessPolicies class.")

        default_action = kwargs.get("defaultAction", {}).get("action")
        if default_action:
            self.defaultAction = default_action

        default_action_id = kwargs.get("defaultAction", {}).get("id")
        if default_action_id:
            self.defaultActionId = default_action_id



"""DeviceGroupRecords class."""

from fmcapi.api_objects.apiclasstemplate import APIClassTemplate
from fmcapi.api_objects.device_services.devicerecords import DeviceRecords
from fmcapi.api_objects.device_ha_pair_services.ftddevicehapairs import FTDDeviceHAPairs
import logging


class DeviceGroupRecords(APIClassTemplate):
    """The DeviceGroupRecords Object in the FMC."""

    VALID_JSON_DATA = ["id", "name", "members"]
    VALID_FOR_KWARGS = VALID_JSON_DATA + []
    URL_SUFFIX = "/devicegroups/devicegrouprecords"

    def __init__(self, fmc, **kwargs):
        """
        Initialize DeviceGroupRecords object.

        Set self.type to "DeviceGroup" and parse the kwargs.

        :param fmc (object): FMC object
        :param **kwargs: Any other values passed during instantiation.
        :return: None
        """
        super().__init__(fmc, **kwargs)
        logging.debug("In __init__() for DeviceGroupRecords class.")
        self.type = "DeviceGroup"
        self.parse_kwargs(**kwargs)

    def devices(self, action, members=[]):
        """
        Add/modify name to members field of DeviceGroupRecords object.

        :param action: (str) 'add', 'remove', or 'clear'
        :param membres: (list) List of members in group.
        :return: None
        """
        logging.debug("In devices() for DeviceGroupRecords class.")
        if action == "add":
            for member in members:
                if member["type"] == "device":
                    dev1 = DeviceRecords(fmc=self.fmc)
                    dev1.get(name=member["name"])
                elif member["type"] == "deviceHAPair":
                    dev1 = FTDDeviceHAPairs(fmc=self.fmc)
                    dev1.get(name=member["name"])
                if "id" in dev1.__dict__:
                    if "members" in self.__dict__:
                        self.members.append(
                            {"id": dev1.id, "type": dev1.type, "name": dev1.name}
                        )
                    else:
                        self.members = [
                            {"id": dev1.id, "type": dev1.type, "name": dev1.name}
                        ]
                    logging.info(
                        f'DeviceRecord "{dev1.name}" added to this DeviceGroupRecords object.'
                    )
                else:
                    logging.warning(
                        f"{member} not found.  Cannot add DeviceRecord to DeviceGroupRecords."
                    )
        elif action == "remove":
            if "members" in self.__dict__:
                for member in members:
                    if member["type"] == "device":
                        dev1 = DeviceRecords(fmc=self.fmc)
                        dev1.get(name=member["name"])
                    elif member["type"] == "deviceHAPair":
                        dev1 = FTDDeviceHAPairs(fmc=self.fmc)
                        dev1.get(name=member["name"])
                    if "id" in dev1.__dict__:
                        if member["type"] == "device":
                            self.members = list(
                                filter(lambda i: i["id"] != dev1.id, self.members)
                            )
                        elif member["type"] == "deviceHAPair":
                            devHA1 = FTDDeviceHAPairs(fmc=self.fmc)
                            devHA1.get(name=member["name"])
                            self.members = list(
                                filter(
                                    lambda i: i["id"] != devHA1.primary["id"],
                                    self.members,
                                )
                            )
                            self.members = list(
                                filter(
                                    lambda i: i["id"] != devHA1.secondary["id"],
                                    self.members,
                                )
                            )
                    else:
                        logging.warning(
                            f"DeviceRecord {member} not registered.  Cannot remove DeviceRecord"
                            f" from DeviceGroupRecords."
                        )
            else:
                logging.warning(
                    "DeviceGroupRecords has no members.  Cannot remove DeviceRecord."
                )
        elif action == "clear":
            if "members" in self.__dict__:
                del self.members
                logging.info(
                    "All device records removed from this DeviceGroupRecords object."
                )



"""Sub-interfaces Classes."""

from fmcapi.api_objects.apiclasstemplate import APIClassTemplate
from fmcapi.api_objects.device_services.devicerecords import DeviceRecords
from fmcapi.api_objects.object_services.securityzones import SecurityZones
from fmcapi.api_objects.device_services.physicalinterfaces import PhysicalInterfaces
from fmcapi.api_objects.device_services.etherchannelinterfaces import (
    EtherchannelInterfaces,
)
import logging


class SubInterfaces(APIClassTemplate):
    """The Subinterface Object in the FMC."""

    VALID_JSON_DATA = [
        "id",
        "name",
        "type",
        "mode",
        "enabled",
        "MTU",
        "managementOnly",
        "ipAddress",
        "subIntfId",
        "vlanId",
        "macLearn",
        "ifname",
        "securityZone",
        "arpConfig",
        "ipv4",
        "ipv6",
        "macTable",
        "enableAntiSpoofing",
        "fragmentReassembly",
        "enableDNSLookup",
        "activeMACAddress",
        "standbyMACAddress",
    ]
    VALID_FOR_KWARGS = VALID_JSON_DATA + ["device_name"]
    VALID_CHARACTERS_FOR_NAME = """[.\w\d_\-\/\. ]"""
    PREFIX_URL = "/devices/devicerecords"
    URL_SUFFIX = None
    REQUIRED_FOR_POST = ["name", "subIntfId", "MTU"]
    REQUIRED_FOR_PUT = ["id", "device_id"]
    VALID_FOR_IPV4 = ["static", "dhcp", "pppoe"]
    VALID_FOR_MODE = ["INLINE", "PASSIVE", "TAP", "ERSPAN", "NONE"]
    VALID_FOR_MTU = range(64, 9085)

    def __init__(self, fmc, **kwargs):
        """
        Initialize SubInterfaces object.

        Set self.type to "SubInterface" and parse the kwargs.

        :param fmc (object): FMC object
        :param **kwargs: Any other values passed during instantiation.
        :return: None
        """
        super().__init__(fmc, **kwargs)
        logging.debug("In __init__() for SubInterfaces class.")
        self.parse_kwargs(**kwargs)
        self.type = "SubInterface"

    def parse_kwargs(self, **kwargs):
        """
        Parse the kwargs and set self variables to match.

        :return: None
        """
        super().parse_kwargs(**kwargs)
        logging.debug("In parse_kwargs() for SubInterfaces class.")
        if "device_name" in kwargs:
            self.device(device_name=kwargs["device_name"])
        if "ipv4" in kwargs:
            if list(kwargs["ipv4"].keys())[0] in self.VALID_FOR_IPV4:
                self.ipv4 = kwargs["ipv4"]
            else:
                logging.warning(
                    f"""Method "{kwargs['ipv4']}" is not a valid ipv4 type."""
                )
        if "mode" in kwargs:
            if kwargs["mode"] in self.VALID_FOR_MODE:
                self.mode = kwargs["mode"]
            else:
                logging.warning(f"""Mode "{kwargs['mode']}" is not a valid mode.""")
        if "MTU" in kwargs:
            if kwargs["MTU"] in self.VALID_FOR_MTU:
                self.MTU = kwargs["MTU"]
            else:
                logging.warning(
                    f"""MTU "{kwargs['MTU']}" should be in the range 64-9000."""
                )
                self.MTU = 1500

    def device(self, device_name):
        """
        Associate device to this subinterface.

        :param device_name: (str) Name of device.
        :return: None
        """
        logging.debug("In device() for SubInterfaces class.")
        device1 = DeviceRecords(fmc=self.fmc)
        device1.get(name=device_name)
        if "id" in device1.__dict__:
            self.device_id = device1.id
            self.URL = f"{self.fmc.configuration_url}{self.PREFIX_URL}/{self.device_id}/subinterfaces"
            self.device_added_to_url = True
        else:
            logging.warning(
                f'Device "{device_name}" not found.  Cannot set up device for SubInterfaces.'
            )

    def sz(self, name):
        """
        Assign Security Zone to this subinterface.

        :param name: (str) Name of Security Zone.
        :return: None
        """
        logging.debug("In sz() for SubInterfaces class.")
        sz = SecurityZones(fmc=self.fmc)
        sz.get(name=name)
        if "id" in sz.__dict__:
            new_zone = {"name": sz.name, "id": sz.id, "type": sz.type}
            self.securityZone = new_zone
        else:
            logging.warning(
                f'Security Zone, "{name}", not found.  Cannot add to SubInterfaces.'
            )

    def static(self, ipv4addr, ipv4mask):
        """
        Assign static IP to this bridge subinterface.

        :param ipv4addr: (str) x.x.x.x
        :param ipv4mask: (str) bitmask
        :return: None
        """
        logging.debug("In static() for SubInterfaces class.")
        self.ipv4 = {"static": {"address": ipv4addr, "netmask": ipv4mask}}

    def dhcp(self, enableDefault=True, routeMetric=1):
        """
        Configure this subinterface with DHCP for addressing.

        :param enableDefault: (bool) Accept, or not, a default route via DHCP.
        :param routeMetric: (int) Set route metric.
        :return: None
        """
        logging.debug("In dhcp() for SubInterfaces class.")
        self.ipv4 = {
            "dhcp": {
                "enableDefaultRouteDHCP": enableDefault,
                "dhcpRouteMetric": routeMetric,
            }
        }

    def p_interface(self, p_interface, device_name):
        """
        Define which physical interface on which device is a part of this subinterface.

        :param p_interfaces: (str) Name of physical interface.
        :param device_name: (str) Name of device with that interface.
        :return: None
        """
        logging.debug("In p_interface() for SubInterfaces class.")
        intf1 = PhysicalInterfaces(fmc=self.fmc)
        intf1.get(name=p_interface, device_name=device_name)
        if "id" in intf1.__dict__:
            self.name = intf1.name
            if "MTU" not in self.__dict__:
                self.MTU = intf1.MTU
        else:
            logging.warning(
                f'PhysicalInterface, "{intf1.name}", not found.  Cannot add to SubInterfaces.'
            )

    def e_interface(self, e_interface, device_name):
        """
        Define which etherchannel interface on which device is a part of this subinterface.

        :param e_interfaces: (str) Name of etherchannel interface.
        :param device_name: (str) Name of device with that interface.
        :return: None
        """
        logging.debug("In e_interface() for SubInterfaces class.")
        intf1 = EtherchannelInterfaces(fmc=self.fmc)
        intf1.get(name=e_interface, device_name=device_name)
        if "id" in intf1.__dict__:
            self.name = intf1.name
            if "MTU" not in self.__dict__:
                self.MTU = intf1.MTU
        else:
            logging.warning(
                f'EtherChannelInterface, "{intf1.name}", not found.  Cannot add to SubInterfaces.'
            )



"""Moving the fmc.deployabledevices to an actual api_object."""

import logging
import time


class DeployableDevices(
    object
):  # Can't import APIClassTemplate due to dependency loop.
    """
    Collect a list of FMC managed devices whose configuration is not up-to-date.

    :return: List of devices needing updates.
    """

    URL_SUFFIX = "/deployment/deployabledevices?expanded=true"

    def __init__(self, fmc):
        """
        Initialize DeployableDevices object.

        :param fmc (object): FMC object
        :return: None
        """
        logging.debug("In __init__ for DeployableDevices() class.")

        self.fmc = fmc
        logging.info(
            f"Waiting {self.fmc.wait_time} seconds to allow the FMC to update the list of deployable devices."
        )
        time.sleep(self.fmc.wait_time)
        self.URL = f"{self.fmc.configuration_url}{self.URL_SUFFIX}"

    def get(self):
        """
        Use GET API call to query FMC for a list of devices that need configuration updates pushed to them.

        :return: (list) uuids
        """
        logging.debug("GET method for API for DeployableDevices.")
        logging.info("Getting a list of deployable devices.")
        response = self.fmc.send_to_api(method="get", url=self.URL)
        # Now to parse the response list to get the UUIDs of each device.
        if "items" not in response:
            return
        uuids = []
        for item in response["items"]:
            if not item["canBeDeployed"]:
                pass
            else:
                uuids.append(item)
        return uuids

    def post(self):
        """POST method for API for DeployableDevices not supported."""
        logging.info("POST method for API for DeployableDevices not supported.")
        pass

    def put(self):
        """PUT method for API for DeployableDevices not supported."""
        logging.info("PUT method for API for DeployableDevices not supported.")
        pass

    def delete(self):
        """DELETE method for API for DeployableDevices not supported."""
        logging.info("DELETE method for API for DeployableDevices not supported.")
        pass
